! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!   caltest_om.f90  --- f90 program to test the CAL (OM aspects only)
!
!   DESCRIPTION:
!       This program tests the correctness of the OM-related CAL part by
!       performing a sequence of subroutine calls followed by
!       checks of the output. If any of these checks fails, the
!       program shall exit with a non-zero return value. If all tests
!       succeed 0 will be returned to the calling environment.
!
program caltest_om
use errorhandling
use cal
use caloaldefs

implicit none
    logical                         :: isClose
    type(OmPsfT)                    :: psf
    type(OMframeParameters)         :: frameParameters
    integer(int32)                  :: instrumentOut, ccdChipIdOut, &
                                       ccdNodeIdOut, accLevelOut, i, j
    integer(int8)                   :: nBranch, nValRanges, ii
    real(double)                    :: stdCol12, stdMag2, &
                                       stdCol12e, stdMag2e, mag
    real(kind=double)               :: plateScale
    real(single), dimension(0:3)    :: rawX = (/10.2, 47.2, 456.4, 999.9/), &
                                       rawY = (/20.4, 974.2, 765.9, 1345.9/)
    real(single), dimension(:), pointer :: deltaX0 => null(), deltaY0 => null(), &
                                           deltaX1 => null(), deltaY1 => null()
    real(single), dimension(:,:), pointer  :: largeSenseMap => null()
    real(double)                    :: lffa, lffae, lpixerr
    real(single), dimension(:,:), pointer  :: smallSenseMap => null()
    real(double)                    :: sffa, sffae, spixerr
    real(single), dimension(:,:), pointer  :: ledMap => null()
    real(double)                    :: leda, ledae, ledpixerr
    real(double), dimension(:), pointer  :: lamda => null(), isf => null(), &
                                            isfError => null()
    real(single)                    :: l10, l20, i10, ie10
    type(OmColorTransformatorT)     :: colorTrans
    type(IntervalT)                 :: colorValInt
    type(ValueErrorT)               :: stdCol, stdMag

    type(ValueAsymmetricErrorT)     :: rawRate, naturalRate


    call CAL_setState('om.fits')
    call CAL_getState(instrument=instrumentOut, ccdChipId = ccdChipIdOut, &
                      ccdNodeId=ccdNodeIdOut, accuracyLevel = accLevelOut)
    if (instrumentOut /= OM) &
        call fatal('caltest_om', "instrumentOut /= OM")
    if (ccdChipIdOut /= 1) &
        call fatal('caltest_om', "ccdChipIdOut /= 1")
    if (ccdNodeIdOut /= 0) &
        call fatal('caltest_om', "ccdNodeIdOut /= 0")

    !
    !   color transformation
    !
    call CAL_setCalibrationFile('mycolor.ccf')
    colorTrans = CAL_omGetColorTransformator(UVW1, V)
    nBranch = CAL_omColorTransBranches(colorTrans)
    write(*, *) 'There are/is ', nBranch, ' branche(s) in the C-C'
    call CAL_omColorTransSetBranch(colorTrans, 0_int8)
    nValRanges = CAL_omColorTransValidityRanges(colorTrans)
    write(*, *) 'Validity ranges:'
    write(*, *) 'min  -   max'
    do ii=0, nValRanges-1
       colorValInt = CAL_omColorTransValidityRange(colorTrans, ii)
       write(*, *) 'validity range: ', colorValInt%lower, colorValInt%upper
    enddo

    stdCol = CAL_omStandardColor(colorTrans, 4.5_double, 5.6_double);
    stdMag = CAL_omStandardMagnitude(colorTrans, stdCol, 5.6_double);
    write(*, *) 'std color     : ', stdCol%value, stdCol%error
    write(*, *) 'std magnitue  : ', stdMag%value, stdMag%error

    call CAL_omColorTransform(UVW1, V, 4.5_double, 5.6_double, &
                              stdCol12, stdMag2, stdCol12e, stdMag2e)
    write(*, *) 'standard color/magnitude + errors'
    write(*, *) stdCol12, stdCol12e
    write(*, *) stdMag2, stdMag2e
    call CAL_omColorTransform(U, B, 12.0_double, 12.9_double, &
                              stdCol12, stdMag2, stdCol12e, stdMag2e)
    write(*, *) stdCol12, stdCol12e
    write(*, *) stdMag2, stdMag2e

    !
    !   OM count rate to magnitude conversion
    !
    call CAL_setState(filterId = UVW1)
    mag = CAL_omNaturalMagnitude(10.3_double)
    write(*, *) 'Natural magnitude for UVW1 filter: ', mag

    !
    !   count rate conversion
    !
    rawRate%value = 3.4
    rawRate%errorMinus = 0.
    rawRate%errorPlus = 0.
    write(*, *) 'count rate conversion:'
    naturalRate = CAL_omPhotoNatural(rawRate, .1_double, .01_double, &
                       .01_double, .false., .true.)
    write(*, *) rawRate%value, ' -> ', naturalRate%value
    write(*, *)

    !
    !   EE
    !
    call CAL_setState(filterId = UVW1)
    psf = CAL_omGetPSF(1024._double, 1024._double, .25_double)
    write(*, *) 'EE(r=10): ', CAL_omPsfEncircledEnergy(psf, 10._double)
    write(*, *) 'R(EE=.5): ', CAL_omPsfCircleRadius(psf, .5_double)

    !
    !   OM distortion
    !
    call CAL_omDistortion(rawX, rawY, deltaX0, deltaY0)
    call CAL_omDistortion(rawX, rawY, deltaX1, deltaY1)
    write(*, *) 'rawX / rawY, deltaXY_inter, deltaXY_poly:'
    do i=0, size(rawX)-1
        write(*, '(2F6.2,2X,4F8.3)') rawX(i), rawY(i), &
                                     deltaX0(i), deltaY0(i), &
                                     deltaX1(i), deltaY1(i)
        if (.not. isClose(deltaX0(i), deltaX1(i)) .or. &
            .not. isClose(deltaY0(i), deltaY1(i))) &
            call error('caltest_om: problem with distortion correction')
    enddo

    ! Test filters GRISM10, GRISM20
    call CAL_setState(filterId = GRISM10)
    call CAL_omDistortion(rawX, rawY, deltaX0, deltaY0)
    write(*,*)' Filter: GRISM10'
    i=2
    write(*, '(2F6.2,2X,4F8.3)') rawX(i), rawY(i), &
                                     deltaX0(i), deltaY0(i), &
                                     deltaX1(i), deltaY1(i)
    call CAL_setState(filterId = GRISM20)
    call CAL_omDistortion(rawX, rawY, deltaX0, deltaY0)
    write(*,*)' Filter: GRISM20'
    write(*, '(2F6.2,2X,4F8.3)') rawX(i), rawY(i), &
                                     deltaX0(i), deltaY0(i), &
                                     deltaX1(i), deltaY1(i)

    !
    !   large/small-scale flatfields
    !
    call CAL_setState(filterId = UVW1)
    call CAL_omLargeSenseVariation(1020, 1020, 8, 8, 0._double, &
                                   largeSenseMap, lffa, lffae, lpixerr);
    write(*, *) 'extents of large-sense-map ', size(largeSenseMap, 1), &
                'x', size(largeSenseMap, 2)
    do i=0, 7
        write(*, '(100f6.3)') (largeSenseMap(i, j), j=0, 7)
    enddo
    write(*, *) 'errors: ', lffa, lffae, lpixerr
    call CAL_releaseMemory(largeSenseMap)


    call CAL_omPixelSenseVariation(1020, 1020, 8, 8, 0._double, &
                                   smallSenseMap, sffa, sffae, spixerr);
    write(*, *) 'extents of small-sense-map ', size(smallSenseMap, 1), &
                'x', size(smallSenseMap, 2)
    do i=0, 7
        write(*, '(100f6.3)') (smallSenseMap(i, j), j=0, 7)
    enddo
    write(*, *) 'errors: ', sffa, sffae, lpixerr
    call CAL_releaseMemory(smallSenseMap)

    !
    !   OM LED template
    !
    call CAL_omLEDtemplateMap(1020, 1020, 8, 8, ledMap, leda, ledae, ledPixErr)
    write(*, *) 'extents of LED template map ', size(ledMap, 1), &
                'x', size(ledMap, 2)
    do i=0, 7
        write(*, '(100f6.3)') (ledMap(i, j), j=0, 7)
    enddo
    write(*, *) 'errors: ', leda, ledae, ledPixErr
    call CAL_releaseMemory(ledMap)

    !
    !   OM frame parameters (needed to calculate frame time)
    !
    call CAL_omGetFrameParameters(frameParameters)
    write(*, *) 'frame parameters:'
    write(*, *) 'n horizontal pixels     : ', frameParameters%nHorizontalPixels
    write(*, *) 'n vertical pixels       : ', frameParameters%nVerticalPixels
    write(*, *) 'vertical transfer time  : ', &
        frameParameters%verticalTransferTime
    write(*, *) 'horizontal transfer time: ', &
        frameParameters%horizontalTransferTime
    write(*, *) 'vert image pixels       : ', &
        frameParameters%nVerticalPixelsImage
    write(*, *) 'vert->horiz. switch time: ', &
        frameParameters%verticalHorizontalSwitchTime
    write(*, *) 'horiz->vert. switch time: ', &
        frameParameters%horizontalVerticalSwitchTime
    write(*, *) 'redundant CCD y-offset  : ', &
        frameParameters%redundantCcdOffset
    write(*, *)

    !
    !   OM grism parameters 
    !
    call CAL_setState(filterId = GRISM1)
    call CAL_omGetGrismFlux(lamda, isf, isfError)
    write(*,* )' Grism 1: lamda, isf, isferr',lamda(10), &
                                         isf(10),isfError(10)
    l10 = 2000.0
    i10= 3.018E-12
    ie10=10.062
        if (.not. isClose(real(lamda(10), single), l10) .or. &
            .not. isClose(real(isf(10), single), i10) .or. &
            .not. isClose(real(isfError(10), single), ie10)) &
            call error('caltest_om: problem with Grism flux')

    call CAL_setState(filterId = GRISM2)
    call CAL_omGetGrismWavelength(lamda)
    l20=313.4
    write(*,* )' Grism 2: lamda',lamda(20)
        if (.not. isClose(real(lamda(20), single), l20)) &
            call error('caltest_om: problem with Grism wavelength')

    call message(AppMsg, Verbose, "all tests passed")
end program


logical function isClose(value, trueval)
use types

    real(single), intent(in)    :: value, trueval
    real(single)                :: absval, abstrueval

    isClose = .true.
    if (value * trueval .lt. 0.) isClose = .false.
    absval = abs(value)
    abstrueval = abs(trueval)
    if (absval .lt. .9999 * abstrueval .or. &
        absval .gt. 1.001 * abstrueval) isClose = .false.
end function
