// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalServer.cc	--- implementation of CalServer class methods
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1998-04-22: 1.0.0: - first version
//		UL, 1998-11-18: 2.0.0: - overhaul
//		UL, 1999-02-10: 2.1.0: - use templated member functions
//		UL, 1999-03-29: 3.0.0: - overhaul in line with CalServer.h-3.0.0
//		UL, 1999-04-12: 3.1.0: - allow ERM as valid instrument
//		UL, 1999-04-22: 3.1.1: - ugly change: allow different units of FRMTIME
//								 attribute in recognition of EMOS mode
//		UL, 1999-05-12: 3.2.2: - changed fatal -> error and changed all
//								 error/warning tags
//		UL, 1999-09-10: 3.3.0: - added AbsorptionDataServer
//		UL, 2000-01-05: 3.4.0: - retrieval of atoms produces now proper error
//								 message in case of missing CCF data
//		UL, 2000-05-24: 3.5.0: - for figuring out the camera mode read the
//								 CCF mode parameters and pass data to
//								 EMOS/EPNcam::toMode()
//		UL, 2000-09-28: 2.6.0: - setState(): also set the filter variable
//								 (from "FILTER" attribute)
//		UL, 2000-10-20: 2.7.0: - fixed SPR 1667: MOS RFS mode: Outer CCDs are
//								 not in RFS (see corresponding code comment in
//								 CalServer::setState())
//		UL, 2000-11-10: 2.7.1: - some minor changes to eliminate warnings
//								 from optimizer
//		UL, 2001-07-18: 2.7.2: - also allow FILTER="UNDEFINED" in definition
//								 of state
//		UL, 2001-11-26: 2.8.0: - if SUBMODE attribute is present determine
//								 camera mode from value and skip mode
//								 recognition logic
//		UL, 2003-30-01: 2.9.0: - no longer die on DATE-END=-1 in setState()
//							   - supress warning message about missing
//								 CCDID/NODEID in setState() when invoked on
//								 merged event list
//		UL, 2003-05-23: 2.10.0:- CalServer.setState() now honors value of
//								 'silent' argument when setting state
//								 variables
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/CalServer.cc,v 1.86 2013/08/23 11:05:58 rsaxton Exp $
//
#include "CalServer.h"
#include "RealCcf.h"
#include "DalUtils.h"
#include "errstr.h"
#include "Random.h"

using namespace std;

//
//	---( CalServer )-------------------------------------------------------
//
CalServer::CalServer(Xmm::Instrument instrument) :
	_instrument(0), _state(0), _ccf(0), _atomServer(0),
	_oldRandMin(-.5), _oldRandMax(.5)
{
	setInstrument(instrument);
	setRandom();
} // CalServer::CalServer()


void CalServer::setInstrument(Xmm::Instrument instrument)
{
	if (_instrument && _instrument->id() == instrument)
		//	nothing to change
		return;

	init();

	errstr<<"Setting CAL state to instrument "<<instrument<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	switch (instrument) {
		case Xmm::EMOS1: _instrument	= new EMOScam1; break;
		case Xmm::EMOS2: _instrument	= new EMOScam2; break;
		case Xmm::EPN:	 _instrument	= new EPNcam; break;
		case Xmm::ERM:	 _instrument	= new ERMcam; break;
		case Xmm::RGS1:	 _instrument	= new RGScam1; break;
		case Xmm::RGS2:	 _instrument	= new RGScam2; break;
		case Xmm::OM:	 _instrument	= new OMcam; break;
		case Xmm::SC:	 _instrument	= new XmmInstrument; break;
		default		   : errstr<<"illegal instrument `"
							   <<Xmm::toString(instrument)<<"'"
							   <<Sas::error("IllegalInstrument");
	}
	defState();
} // CalServer::setInstrument()


CalState &CalServer::getState() const
{
	return *_state;
} // CalServer::getState();


void CalServer::setState(const CalState &calState)
{
	delete _state;
	_state = new CalState(calState);
} // CalServer::setState()


void CalServer::pushState()
{
	_stateStack.push_back(_state);
	defState();
} // CalServer::pushState()


void CalServer::popState()
{
	if (_stateStack.size()) {
		delete _state;
		_state = _stateStack[_stateStack.size() - 1];
		_stateStack.pop_back();
	} else
		errstr<<"Cannot pop state stack because it is empty - state remains "
				"unchanged"
			  <<Sas::error("OperationNotPermitted");
} // CalServer::pushState()


CalState *CalServer::state() const
{
	return _state;
} // CalServer::state()


Ccf *CalServer::setCcf(const string &cifName) const
{
	delete _ccf;
	delete _atomServer;
	_ccf = new RealCcf(cifName);
	_atomServer = new CalAtomCollectionServer(this);

	return _ccf;
} // CalServer::setCcf()


Ccf *CalServer::ccf() const
{
	checkCcf();

	return _ccf;
} // CalServer::ccf();


void CalServer::setRandom(double min, double max)
{
	errstr<<"Setting interval for random number generator to ["
		  <<min<<", "<<max<<"]\n"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	_oldRandMin	= randGen.min();
	_oldRandMax = randGen.max();
	randGen.minMaxRange(min, max, randGen.rangeType());
} // CalServer::setRandom()


void CalServer::setRandom()
{
	randGen.minMaxRange(_oldRandMin, _oldRandMax, randGen.rangeType());
} // CalServer::setRandom()


CalAtom *CalServer::getAtom(Ccf::CalDataTypes)
{
	errstr<<"Not implemented."<<Sas::warning("NotYetImplemeneted");

	checkCcf();
	static BadPixelDataServer *map = getAtom(map);

	return map;
} // CalServer::getAtom()


//
//	following templated member function needs to be specialized for
//	each new CAL atom that is added; there is convenience macro
//	getAtomImp() defined below to ease this
//	If a new CalAtom X is defined a new getAtomImp(X) needs to to be added
//	below
//
template<class AtomType>
AtomType *CalServer::getAtom(AtomType *p, bool generateError) const
{
	checkCcf();

	if (errHandler.verbosity(Msg::Lib) >= Msg::Verbose)
		errstr<<"Retrieving CAL atom "<<Xmm::toString(_instrument->id())<<"/"
			  <<Ccf::toString(p->id())<<" corresponding to "
			  <<"CCD  = "<<(_state->ccd()->isActive() ?
							toString(_state->ccd()->get()) : "*inactive*")
						 <<endl
			  <<"node = "<<(_state->node()->isActive() ?
						Xmm::toString(_state->node()->get()) : "*inactive*")
						 <<endl
			  <<"mode = "<<(_state->mode()->isActive() ?
						Xmm::toString(_state->mode()->get()) : "*inactive*")
						 <<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

	CalAtomCollection<AtomType> *ptr =
								_atomServer->getCollection(_instrument, p);

	AtomType *atom = ptr->select(
		_state->ccd()->isActive() ? _state->ccd()->get() : 0,
		_state->node()->isActive() ? _state->node()->get() : Xmm::PRIMARY_NODE,
		_state->mode()->isActive() ?
							_instrument->modeIndex(_state->mode()->get()) : 0);

	if (!atom && generateError)
		errstr<<"The CCF constituent "<<Xmm::toString(_instrument->id())<<"/"
			  <<Ccf::toString(p->id())<<" does not contain an entry "
				"corresponding to "
			  <<"CCD="<<(_state->ccd()->isActive() ?
							toString(_state->ccd()->get()) : "*inactive*")
			  <<"/node="<<(_state->node()->isActive() ?
						Xmm::toString(_state->node()->get()) : "*inactive*")
			  <<"/mode="<<(_state->mode()->isActive() ?
						Xmm::toString(_state->mode()->get()) : "*inactive*")
						 <<endl
			  <<Sas::error("NoSuchEntry");

	return atom;
} // CalServer::getAom()

#define getAtomImp(a)	template a *CalServer::getAtom(a *, bool) const;

getAtomImp(AbsorptionDataServer)
getAtomImp(BackgroundDataServer)
getAtomImp(BadPixelDataServer)
getAtomImp(Boresight)
getAtomImp(CanonicalCrossPsf)
getAtomImp(CanonicalRedist)
getAtomImp(CcdQuantumDataServer)
getAtomImp(ContaminantDataServer)
getAtomImp(CtiCorrector)
getAtomImp(DarkFrameDataServer)
getAtomImp(EffectiveAreaDataServer)
getAtomImp(EncircledEnergyDataServer)
getAtomImp(EnergyCorrector)
getAtomImp(FilterDataServer)
getAtomImp(FWCDataServer)
getAtomImp(GeometryDataServer)
getAtomImp(GratingDataServer)
getAtomImp(HKvalRangesServer)
getAtomImp(LowEnergyNoiseServer)
getAtomImp(MiscDataServer)
getAtomImp(ModeParameters)
getAtomImp(OmColorTransformator)
getAtomImp(OmGrismDataServer)
getAtomImp(OmFlatFieldServer)
getAtomImp(OmFluxConverter)
getAtomImp(OmGeoLinearizer)
getAtomImp(OmPsf)
getAtomImp(PatternLibServer)
getAtomImp(PsfDataServer)
getAtomImp(RgsBackgroundServer)
getAtomImp(RgsCoolPixDataServer)
getAtomImp(RgsSolarAspectAngServer)
getAtomImp(ScreenThresholdServer)
getAtomImp(SpecQualityServer)
getAtomImp(TimeCorrector)
getAtomImp(TimeJumpTolServer)

#undef getAtomImp

//
//	templated member function to release atoms from a buffer; as above
//	this needs to be specialized for each new atom that is added; convenience
//	macro relAtomImp() is defined below to ease this
//	If a new CalAtom X is defined a new relAtomImp(X) needs to be added
//	below
//
template<class AtomType>
void CalServer::releaseAtom(AtomType *p) const
{
	checkCcf();
	_atomServer->releaseCollection(_instrument, p);
} // CalServer::releaseAtom()

#define relAtomImp(a)	template void CalServer::releaseAtom(a *) const;

relAtomImp(AbsorptionDataServer)
relAtomImp(BackgroundDataServer)
relAtomImp(BadPixelDataServer)
relAtomImp(Boresight)
relAtomImp(CanonicalCrossPsf)
relAtomImp(CanonicalRedist)
relAtomImp(CcdQuantumDataServer)
relAtomImp(ContaminantDataServer)
relAtomImp(GeometryDataServer)
relAtomImp(CtiCorrector)
relAtomImp(DarkFrameDataServer)
relAtomImp(EffectiveAreaDataServer)
relAtomImp(EncircledEnergyDataServer)
relAtomImp(EnergyCorrector)
relAtomImp(FilterDataServer)
relAtomImp(GratingDataServer)
relAtomImp(HKvalRangesServer)
relAtomImp(LowEnergyNoiseServer)
relAtomImp(MiscDataServer)
relAtomImp(ModeParameters)
relAtomImp(OmColorTransformator)
relAtomImp(OmGrismDataServer)
relAtomImp(OmFlatFieldServer)
relAtomImp(OmFluxConverter)
relAtomImp(OmGeoLinearizer)
relAtomImp(OmPsf)
relAtomImp(PatternLibServer)
relAtomImp(PsfDataServer)
relAtomImp(RgsBackgroundServer)
relAtomImp(RgsCoolPixDataServer)
relAtomImp(RgsSolarAspectAngServer)
relAtomImp(ScreenThresholdServer)
relAtomImp(SpecQualityServer)
relAtomImp(TimeCorrector)
relAtomImp(TimeJumpTolServer)

#undef relAtomImp


CalAtomCollectionServer *CalServer::atomServer() const
{
	checkCcf();

	return _atomServer;
} // CalServer::atomServer()


//
//	Print information about the server settings (instrument + state)
//	on given ostream
//
void CalServer::print(ostream &os) const
{
	os<<"   instrument             : "<<UNT(_instrument->id())<<
				"["<<Xmm::toString(_instrument->id())<<"]\n"<<
		*_state<<endl;
} // CalServer::print()


CalInt16Vector CalServer::int16Vector(const int16 *ptr, unsigned n)
{
	return CalInt16Vector(const_cast<int16 *>(ptr), n, this);
} // CalServer::int16Vector()


CalReal32Vector CalServer::real32Vector(const real32 *ptr, unsigned n)
{
	return CalReal32Vector(const_cast<real32 *>(ptr), n, this);
} // CalServer::real32Vector()


CalServer::~CalServer()
{

	  // FIXME: at this time ~CalServer is never executed in the SAS.
	  if(ccfAccessLog().size()){
	  errstr<<"CCF constituents accessed by the calibration server:\n"
			<<ccfAccessLog()<<endl
			<<Sas::message(Msg::Lib, Msg::Always);
	  // Re-issue warnings about expired CCF constituents, if any.
	  ccfAccessLog().evaluate();
	  }

	init();
	delete _atomServer;
	delete _ccf;
	//	restore RNG to what it was
	randGen.minMaxRange(_oldRandMin, _oldRandMax, randGen.rangeType());
} // CalServer::~CalServer()


//
//	Construct a default state
//
void CalServer::defState()
{
	switch (_instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2: _state			= new CalStateEmos(this);
						 break;
		case Xmm::EPN:	 _state			= new CalStateEpn(this);
						 break;
		case Xmm::RGS1:
		case Xmm::RGS2:	 _state			= new CalStateRgs(this);
						 break;
		case Xmm::OM:	 _state			= new CalStateOm(this);
						 break;
		case Xmm::ERM:
		case Xmm::SC:	 _state			= new CalState(this);
						 break;

		default		   : errstr<<"illegal instrument `"
							   <<Xmm::toString(_instrument->id())<<"'"
							   <<Sas::error("IllegalInstrument");
	}
} // CalServer::defState()


void CalServer::init()
{
	delete _instrument;	_instrument = 0;
	delete _state; _state = 0;
	while (_stateStack.size()) {
		delete _stateStack[_stateStack.size()-1];
		_stateStack.pop_back();
	}
} // CalServer::init()


void CalServer::checkCcf() const
{
	if (!_ccf)
		//	no Ccf defined yet - instantiate one first
		setCcf();
} // CalServer::checkCcf()


vector<CoreModeParameters> CalServer::coreModeParameters()
{
	pushState();
	vector<CoreModeParameters> v;
	const vector<Mode> &modes = _instrument->modes();
	ModeParameters *modeParams = 0;
	for (unsigned n=0; n<modes.size(); ++n) {
		_state->mode()->set(modes[n]);
		/* AI gcc4 porting
		 * 'modeParams' may be used uninitialized in this function =>
		 * Has to be defined outside the for loop??
		 */
		modeParams = getAtom(modeParams, false);
		if (!modeParams)
			break;
		CoreModeParameters cmp;
		cmp.windowX0 = modeParams->windowX0();
		cmp.windowY0 = modeParams->windowY0();
		cmp.windowDX = modeParams->windowDX();
		cmp.windowDY = modeParams->windowDY();
		cmp.frameTime = UNT(modeParams->frameTime());
		v.push_back(cmp);
	}
	popState();

	return v;
} // CalServer::coreModeParameters()

namespace {
ostream &sciwarn(const string &listOfAttribs,
				 const string &statevar, bool silent)
{
	if (!silent)
		errstr<<"In defining the Cal state from a data set or block "
				"(one or more of) the mandatory attribute(s) "
			  <<listOfAttribs
			  <<" was/were not found - from this/these attribute(s) the "
			  <<"actual "<<statevar
			  <<" is derived which is now not possible - will "
				"assume a default value for the "<<statevar<<" - please check "
				"the diagnostic output below (requires verbosity >= Verbose) "
				"for the effective Cal state settings. Please note: "
				"In case of a possibly incorrect "<<statevar<<" the retrieval "
				"of "<<statevar<<"-dependent calibration quantities via the "
				"Cal is likely to yield INADEQUATE data which ultimately might"
				" lead to INCORRECT scientific results"
			  <<Sas::warning("MandatoryAttributeNotFound");

	return errstr;
} // sciwarn()
}

//
//	Following setState() is the most fundamental of the setState() methods;
//	defines the CAL state from a vector of Dal::Attributables (usually
//	from an event list) and the event list itself;
//	ARGUMENTS:
//		av		: vector with attributes from which to define the state
//		silent	: if needed attributes are not found in the passed Attributable
//				  vector a warning is raised and the corresponding state
//				  variable is set to a default value; setting silent=false
//				  will suppress the warning message (needed e.g. in odfingest)
//		ignoreFilter:
//				  if set to true the filter state variable will not be
//				  attempted to set from a corresponding attribute
//		evTable	: pointer to event table; needed if CCDID is not found
//				  among attribute list (merged event list)
//
void CalServer::setState(const vector<const Attributable *> &av,
						 bool silent, bool ignoreFilter, const Table *evTable)
{
	//
	//	see if we have an XMM data set
	//
	try {
		string s(DalUtils::attribute(av, "TELESCOP")->asString());
		if (s != "XMM")
			errstr<<"The data set used to define the Cal state does not seem "
					"to be an XMM set - the value of the attribute "
					"`TELESCOP\' is: "<<s
				  <<Sas::error("IllegalAttributeValue");
	} catch(...) {
		errstr<<"No attribute named `TELESCOP\' found"
			  <<Sas::error("MandatoryAttributeNotFound");
	}

	//
	//	determine the instrument
	//
	Xmm::Instrument instr = Xmm::EMOS1;
	try {
		instr = Xmm::toInstrument(DalUtils::attribute(av,
													"INSTRUME")->asString());
	} catch(...) {
		errstr<<"The data set used to define the Cal state does not contain "
				"an attribute named `INSTRUME\' - this is needed to determine "
				"the current instrument."
			  <<Sas::error("MandatoryAttributeNotFound");
	}
	setInstrument(instr);

	//
	//	at this point, the state variables have their default values
	//
	CalState *state = this->state();

	//
	//	check if there is SUBMODE
	//
	bool modeSet = false;
	try {
		state->mode()->set(Xmm::toMode(
							DalUtils::attribute(av, "SUBMODE")->asString()));
		modeSet = true;
	} catch(...) {}

	//	instrument-specific state settings
	switch(instr) {
		case Xmm::EMOS1:
		case Xmm::EMOS2:
			try {
				state->ccd()->set(DalUtils::attribute(av, "CCDID")->asInt());
			} catch(...) {
				//	no CCDID - check if there is a CCDNR column; if there
				//	is this is a merged event list and CCDID is not applicable
				if (evTable && !evTable->hasColumn("CCDNR"))
					sciwarn("`CCDID\'", "CCD", silent);
			}
			try {
				state->node()->
					set(Xmm::toNode(DalUtils::attribute(av,
														"CCDNODE")->asInt()));
			} catch(...) {
				//	no CCDID - check if there is a CCDNR column; if there
				//	is this is a merged event list and CCDID is not applicable
				if (evTable && !evTable->hasColumn("CCDNR"))
					sciwarn("`CCDNODE'", "node", silent);
			}

			if (!modeSet) {
				pair<Mode, bool> mode = EMOScam::toMode(av,
												coreModeParameters(),
												state->ccd()->get(),
												silent);
				if (state->ccd()->get() > 1 &&
					mode.first == Xmm::PRIME_PARTIAL_RFS)
					//
					//	For camera-RFS mode the inner CCD is operated
					//	in Xmm::PRIME_PARTIAL_RFS and the outer ones
					//	in Xmm::PRIME_FULL_WINDOW mode; however, the
					//	ODS for all CCDs wrongly puts the RFS frame
					//	time in the event list, hence, the mode detection
					//	logic gets fooled and wrongly yields
					//	Xmm::PRIME_PARTIAL_RFS for the outer CCDs
					//
					mode.first = Xmm::PRIME_FULL_WINDOW;
				state->mode()->set(mode.first, !mode.second);
			}
			break;
		case Xmm::EPN:
			try {
				state->ccd()->set(EPNcam::toChipNr(
								DalUtils::attribute(av, "CCDID")->asInt(),
								DalUtils::attribute(av, "QUADRANT")->asInt()));
			} catch(...) {
				if (evTable && !evTable->hasColumn("CCDNR"))
				   sciwarn("`CCDID\', `QUADRANT\'", "CCD", silent);
			}

			if (!modeSet) {
				pair<Mode, bool> mode = EPNcam::toMode(av,
													   coreModeParameters(),
													   silent);
				state->mode()->set(mode.first, !mode.second);
			}
			break;
		case Xmm::RGS1:
		case Xmm::RGS2:
			try {
				state->ccd()->set(DalUtils::attribute(av, "CCDID")->asInt());
			} catch (...) {
				if (evTable && !evTable->hasColumn("CCDNR"))
				   sciwarn("`CCDID\'", "CCD", silent);
			}

			try {
				state->ocb()->set(DalUtils::attribute(av, "CCDOCB")->asInt());
			} catch (...) {
				sciwarn("`CCDOCB\'", "OCB", silent);
			}

			if (!modeSet) {
				pair<Mode, bool> mode = RGScam::toMode(av,
													   state->ocb()->get(),
													   silent);
				state->mode()->set(mode.first, !mode.second);
			}
			break;
		case Xmm::OM:
			state->ccd()->set(1);	//	lovely, there is only one CCD
			if (!modeSet) {
				pair<Mode, bool> mode = OMcam::toMode(av, silent);
				state->mode()->set(mode.first, !mode.second);
			}
			break;
		default:
			errstr<<"Initialization from `"<<Xmm::toString(instr)<<"' not "
					"supported"<<Sas::error("IllegalInstrument");
			break;
	}

	//
	//	set filter
	//
	if (state->filter()->isActive()) {
		if (!ignoreFilter)
			try {
				string filter = DalUtils::attribute(av, "FILTER")->asString();
				if (filter.empty() || uppercase(filter) == "UNKNOWN"
								   || uppercase(filter) == "UNDEFINED") {
					errstr<<"Value of FILTER attribute is `"<<filter
						  <<"' - cannot determine active filter"
						  <<Sas::warning("FilterUnknown");
					state->filter()->uncertain();
				} else
					state->filter()->set(Xmm::toFilter(filter));
			} catch(...) {
				if (!silent)
					sciwarn("`FILTER\'", "filter", silent);
				state->filter()->uncertain();
			}
		else
			state->filter()->uncertain();
	}

	//
	//	exposure start time
	//
	try {
		STime::errorHandling = STime::EXCEPTION;
		STime tstart(DalUtils::attribute(av, "DATE-OBS")->asString()),
			  tend(DalUtils::attribute(av, "DATE-END")->asString());
		state->exposureStartTime()->set(tstart, false, silent);
		state->exposureDuration()->set(tend-tstart, false, silent);
	} catch(...) {
		sciwarn("`DATE-OBS/DATE-END\'", "exposure start/end date", silent);
	}

	//
	//	all done, output distagnostic messges
	//
	errstr<<"Cal state set to:\n"
		  <<"   instrument             : "<<instr<<endl
		  <<*state<<Sas::message(Msg::Lib, Msg::Verbose);
} // CalServer::setState()


void CalServer::setState(const string &dataSet, bool silent, bool ignoreFilter)
{
	DataSet *set = dataSetServer->open(dataSet, DataSet::Read);
	setState(*set, silent, ignoreFilter);
	dataSetServer->close(set);
} // CalServer::setState()


void CalServer::setState(const Block &block, bool silent, bool ignoreFilter)
{
	vector<const Attributable *> av;
	av.push_back(&block);
	setState(av, silent, ignoreFilter, block.type() == Block::TableT
										? block.parent()->table(block.name())
										: 0);
} // CalServer::setState()


void CalServer::setState(const DataSet &dataSet, bool silent,
						 bool ignoreFilter)
{
	vector<const Attributable *> av;
	av.push_back(&dataSet);
	const Block *block = dataSet.block(0);
	av.push_back(block);
	setState(av, silent, ignoreFilter, block->type() == Block::TableT
										? block->parent()->table(block->name())
										: 0);
} // CalServer::setState()


//
//	Below operator<<() initialize CalServer from contents of (DAL) DataSet or
//	Block; convenience operators to corresponding explicit setState()
//	methods
//
CalServer &operator<<(CalServer &server, const string &dataSet)
{
	server.setState(dataSet);
	return server;
} // operator<<


CalServer &operator<<(CalServer &server, const Block &block)
{
	server.setState(block);

	return server;
} // CalServer::operator<<


CalServer &operator<<(CalServer &server, const DataSet &dataSet)
{
	server.setState(dataSet);

	return server;
} // CalServer::operator<<


CalServer &operator<<(CalServer &server,
					  const vector<const Attributable *> av)
{
	server.setState(av);
	return server;
} // operator<<


ostream &operator<<(ostream &os, CalServer &server)
{
	server.print(os);

	return os;
} // operator<<


//
//	Function to return the CalServer singleton; CalServer.h defines
//	a convenience macro 'calServer' for the function
//
CalServer &calServerFunc()
{
	static CalServer *calServerInstance = new CalServer;
	return *calServerInstance;
} // calServerFunc()
