// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	ContaminantDataServer --- the ContaminantDataServer class
//
//	DESCRIPTION:
//		A class 'ContaminantDataServer' is defined which provides access to
//		contaminant transmission data for the EPIC MOS/PN contaminants.
//
//	CHANGE HISTORY:
//		UL, 2000-02-01: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Tue Feb  1 17:20:27 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Data Access Handbool, C. Erd et al.
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/ContaminantDataServer.h,v 1.1 2013/08/23 11:08:12 rsaxton Exp $
//

#ifndef CONTAMINANTDATASERVER_H
#define CONTAMINANTDATASERVER_H

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "slatec++.h"
#include "Evaluator.h"

//
//	---( ContaminantTransmissionCurve )-------------------------------------------
//
class ContaminantTransmissionCurve : public Slatec::CubicHermiteInterpolant<double>{
public:
	ContaminantTransmissionCurve(double value)
		: Slatec::CubicHermiteInterpolant<double>(value) {}
	ContaminantTransmissionCurve(const std::vector<double> &eGrid,
							const std::vector<double> &trans)
		: Slatec::CubicHermiteInterpolant<double>(eGrid, trans) {}
	ContaminantTransmissionCurve(){}; // DEPERATE ATTEMPT

	virtual double transmission(double E) const;
		//
		//	PURPOSE:
		//		Evaluate contaminant transmission curve at a single energy <E>
		//
		//	ARGUMENTS:
		//		E	: energy in eV
		//
		//	RETURNS:
		//		Value of contaminant transmission curve - unitless number in
		//		range [0, 1]
		//


	virtual double transmission(double Elo, double Ehi) const;
		//
		//	PURPOSE:
		//		Return weighted integral of contaminant transmission curve between
		//		the	two energies
		//
		//	ARGUMENTS:
		//		Elo, Ehi	: energies in eV between which contaminant transmission
		//					  is to be integrated; Elo < Elo
		//
		//	RETURNS:
		//						 Ehi
		//                 1    /
		//			F = ------- |   dE f(E)
		//              Ehi-Elo /
		//						Elo
		//

	~ContaminantTransmissionCurve() {}
};

//
//	---( ContaminantDataServer )----------------------------------------
//
class ContaminantDataServer : public CalAtom,
	 public virtual CalChild<CalAtomCollection<ContaminantDataServer> > {
public:
	ContaminantDataServer(const std::vector<double> &eGrid,
			  CalAtomCollection<ContaminantDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate ContaminantDataServer object from energy grid vector
		//		and corresponding transmission data
		//

        ContaminantDataServer(const vector<double> &time,
                              const vector<double> &cdepth,
                              const vector<double> &odepth,
                              const vector<double> &fdepth,
                              CalAtomCollection<ContaminantDataServer> *parent =0);
		//
		//	PURPOSE:
		//		Instantiate ContaminantDataServer object from 
                //              vectors of element depths
		//


	virtual void dump(std::ostream &) {}
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	static unsigned id() { return Ccf::Contamination; }
		//
		//	PURPOSE:
		//		Return ID of CalAtom
		//


	// virtual const ContaminantTransmissionCurve &at(int rev = 0) const;
		//
		//	PURPOSE:
		//		Return contaminant transmission curve 
                //              applicable at revolution (rev)
		//
		//	ARGUMENTS:
		//		rev	: revolution number
		//

        ContaminantTransmissionCurve * at(double time = 0);

	virtual ~ContaminantDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	//	data members

        vector<double> _time;
        vector<double> _c_depth, _o_depth, _f_depth;
        vector<double> _henke_inverse_mu_c, _henke_inverse_mu_o,
				     _henke_inverse_mu_f;
        vector<double>     _eGrid;
        ContaminantTransmissionCurve _totalContamination;
	static const ContaminantTransmissionCurve _zeroCurve, _oneCurve;
}; // ContaminantDataServer

#endif
