// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EffectiveAreaDataServer.cc --- implementation of EffectiveAreaDataServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the EffectiveAreaDataServer class
//
//	CHANGE HISTORY:
//		UL, 2000-02-01: 1.0.0: - first version
//		UL, 2000-04-25: 1.1.0: - temporarily disabled instantiation of
//					 low-accuracy EffectiveAreaDataServer
//					 (numerical problems with 2-D polynomial fit
//					 from IDL)
//		UL, 2000-05-09: 1.2.0: - bug fix in
//			 		 EffectiveAreaDataServer::vignettingCurve
//								 for RGS
//		UL, 2000-09-15: 1.2.1: - theta>0 was implicitely assumed in
//					 effectiveAreaCurve()
//		UL, 2001-11-01: 1.3.0: - empirical Ae curve correction for RGS;
//					 driven by number in
//					 RGS?_QUANTUMEF:RGA_EFFAREACORR
//		UL, 2002-02-12: 1.4.0: - MediumAccuracyEffectiveAreaDataServer::
//					 effectiveArea(): interpolate in energy
//					 with fourth-order polynomial
//		UL, 2002-03-14: 1.5.0: - modify Ae curve by interpolated empirical
//					 correction data (instead of supplementing
//					 Ae data vector)
//
//	AUTHOR:
//		U. Lammers (UL), Tue Feb  1 18:24:59 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd, et al.
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/EffectiveAreaDataServer.cc,v 1.22 2014/04/03 13:52:34 rsaxton Exp $
//

#include <cmath>
#include <vector>

#include "Tracer.h"
#include "CalOalUtility.h"
#include "CalServer.h"
#include "EffectiveAreaDataServer.h"
#include "GratingDataServer.h"
#include "Interpolator.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( EffectiveAreaCurve )-----------------------------------------------
//
double EffectiveAreaCurve::area(double E) const
{
  return (*this)(E);
} // EffectiveAreaCurve::area()


double EffectiveAreaCurve::area(double Elo, double Ehi) const
{
  if (fabs(Ehi-Elo) < 1.e-8)
    return this->area(Elo);

  return this->integral(Elo, Ehi)/(Ehi-Elo);
} // EffectiveAreaCurve::area()


//
//	---( EffectiveAreaDataServer )-------------------------------------------
//
EffectiveAreaDataServer::EffectiveAreaDataServer(
						 CalAtomCollection<EffectiveAreaDataServer> *parent)
  : CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent)
{
} // EffectiveAreaDataServer::EffectiveAreaDataServer()


void EffectiveAreaDataServer::dump(ostream &)
{
  errstr<<"Sorry dumping EffectiveAreaDataServer not yet implemented"
	<<Sas::warning("NotYetImplemented");
} // EffectiveAreaDataServer::dump()


EffectiveAreaCurve EffectiveAreaDataServer::effectiveAreaCurveFull(
								   double theta,
								   double phi) const
{
  return effectiveAreaCurve(theta, phi);
} //  EffectiveAreaDataServer::effectiveAreaCurveFull()


EffectiveAreaCurve EffectiveAreaDataServer::effectiveAreaCurve(double theta,
							       double phi) const
{
  errstr<<"Generic EffectiveAreaDataServer cannot provide effective area "
    "curve"<<Sas::error("MeaningLessRequest");

  return effectiveAreaCurve(theta, phi);
} //  EffectiveAreaDataServer::effectiveAreaCurveFull()


double EffectiveAreaDataServer::effectiveArea(double E,
					      double theta, double phi) const
{
  return effectiveAreaCurve(theta, phi)(E);
} // EffectiveAreaDataServer::effectiveArea()



EffectiveAreaCurve EffectiveAreaDataServer::effectiveAreaCurve(int,
							       double theta,
							       double phi) const
{
  errstr<<"Generic EffectiveAreaDataServer cannot provide RGS-specific "
    "effective area curve"<<Sas::error("MeaninglessRequest");

  return effectiveAreaCurve(theta, phi);
} // EffectiveAreaDataServer::effectiveAreaCurve()


EffectiveAreaCurve EffectiveAreaDataServer::intrinsicEffectiveAreaCurve(
									int order,
									double theta,
									double phi) const
{
  return effectiveAreaCurve(order, theta, phi);
} // EffectiveAreaDataServer::effectiveAreaCurve()

/*
 * AI: New method to deal with the new Effective Area Correction based on
 * Chebyshev polinomials
 */
EffectiveAreaCurve EffectiveAreaDataServer::realisticEffectiveAreaCurve(
									int order,
									double theta,
									double phi) const
{
  return effectiveAreaCurve(order, theta, phi);
} // EffectiveAreaDataServer::effectiveAreaCurve()

/*
 * EffectiveAreaRectification - overridden in RgsEffAreaServer
 */
EffectiveAreaCurve EffectiveAreaDataServer::effectiveAreaRectification(
                                 EffectiveAreaCurve eain) const
{
     return eain;
}

/*
 * Effective Area Correction between RGS1 and RGS2
 */

std::vector<double>   EffectiveAreaDataServer::effectiveAreaCorrection(int order ) const
{
	std::vector<double> v(1,0);

	return v;
}

std::vector<double>   EffectiveAreaDataServer::effectiveAreaCorrectionEner(int order) const
{
	std::vector<double> v(1,0);

	return v;
}
/*
 * applyAreaCorr - overridden in MediumEffAreaServer
 */
EffectiveAreaCurve EffectiveAreaDataServer::applyAreaCorr(
                                 EffectiveAreaCurve eain) const
{
     return eain;
}

vector<vector<double> > EffectiveAreaDataServer::epochInterpolation(vector<double>,
		vector<vector<double> >,vector<vector<double> > corrVals )
{
	return corrVals;
}

VignettingCurve EffectiveAreaDataServer::vignettingCurve(double theta,
							 double phi) const
{
  VignettingCurve vc(effectiveAreaCurve(theta, phi));
  vc /= effectiveAreaCurve(0., 0.);

  return vc;
} // EffectiveAreaDataServer::vignettingCurve()


VignettingCurve EffectiveAreaDataServer::vignettingCurve(int spectralOrder,
							 double theta,
							 double phi) const
{
  VignettingCurve vc(effectiveAreaCurve(spectralOrder, theta, phi));
  vc /= effectiveAreaCurve(spectralOrder, 0., 0.);

  return vc;
} // EffectiveAreaDataServer::vignettingCurve()

EffectiveAreaDataServer::~EffectiveAreaDataServer()
{
} // EffectiveAreaDataServer::~EffectiveAreaDataServer()


//
//	---( LowAccuracyEffectiveAreaDataServer )----------------------------------
//
LowAccuracyEffectiveAreaDataServer::LowAccuracyEffectiveAreaDataServer(
								       Proxy<EffectiveAreaCurve>
								       onAxisEffectiveAreaCurve,
								       const std::vector<double> &eGrid,
								       const std::vector<DImage> &coefs,
								       double deltaF,
								       double deltaY,
								       double deltaZ,
								       CalAtomCollection<EffectiveAreaDataServer> *parent)
  : CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
    EffectiveAreaDataServer(parent),
    _onAxisEffectiveAreaCurve(onAxisEffectiveAreaCurve),
    _eGrid(eGrid),
    _deltaF(deltaF), _deltaY(deltaY), _deltaZ(deltaZ)
{
  MiscDataServer *mds = 0; mds = calServer.getAtom(mds);
  _focalLength = mds->get("FOCAL_LENGTH");

  //	compute coefficient image
  unsigned N = _eGrid.size();
  if (N != coefs.size())
    errstr<<"Size mismatch between energy grid and coefficient image "
      "vector"
	  <<Sas::error("SizeMismatch");
  if (N < 2)
    errstr<<"Need at least two elements in energy grid vector"
	  <<Sas::error("TooLittleDataPoints");

  unsigned jMax = coefs[0].xExtent(),
    kMax = coefs[0].yExtent();

  _coefsCurve = Image<Slatec::CubicHermiteInterpolant<double> *>(jMax, kMax);
  for (unsigned j=0; j<jMax; ++j)
    for (unsigned k=0; k<kMax; ++k) {
      std::vector<double> v(N);
      for (unsigned n=0; n<N; ++n)
	v[n] = coefs[n][j][k];
      _coefsCurve[j][k] =
	new Slatec::CubicHermiteInterpolant<double>(_eGrid, v);
    }
} // LowAccuracyEffectiveAreaDataServer::LowAccuracyEffectiveAreaDataServer()


inline pair<double, double> LowAccuracyEffectiveAreaDataServer::yz(
								   double theta, double phi)
  const
{
  double	c = _focalLength*(1+_deltaF/_focalLength),
    tanTheta = tan(theta/3600. * DEG_TO_RAD);

  return make_pair(c * tanTheta * cos(phi) + _deltaY,
		   c * tanTheta * sin(phi) + _deltaZ);
} // LowAccuracyEffectiveAreaDataServer::yz()


inline double LowAccuracyEffectiveAreaDataServer::vignettingFactor(
								   double E, const pair<double, double> &yz,
								   unsigned jMax, unsigned kMax) const
{
  double ypow = 1.,
    zpow = 1.,
    v = 0.;

  for (unsigned j=0; j<jMax; ++j) {
    for (unsigned k=0; k<kMax; ++k) {
      v += (*_coefsCurve[j][k])(E) * ypow * zpow;
      ypow *= yz.first;
    }
    zpow *= yz.second;
  }

  return v;
} // LowAccuracyEffectiveAreaDataServer::vignettingFactor()


EffectiveAreaCurve LowAccuracyEffectiveAreaDataServer::effectiveAreaCurve(
									  double theta,	//	[arcsec]
									  double phi)		//	[rad]
  const
{
  pair<double, double> yz = this->yz(theta, phi);
  EffectiveAreaCurve ae(*_onAxisEffectiveAreaCurve);
  const vector<double> &E = ae.abscissa();
  unsigned N = E.size();
  vector<double> v(N);
  unsigned jMax = _coefsCurve.xExtent(),
    kMax = _coefsCurve.yExtent();

  for (unsigned n=0; n<N; ++n)
    v[n] = vignettingFactor(E[n], yz, jMax, kMax);

  return ae*Slatec::CubicHermiteInterpolant<double>(E, v);
} // LowAccuracyDataServer::effectiveAreaCurve()


double LowAccuracyEffectiveAreaDataServer::effectiveArea(double E,
							 double theta,
							 double phi) const
{
  unsigned jMax = _coefsCurve.xExtent(),
    kMax = _coefsCurve.yExtent();
  return (*_onAxisEffectiveAreaCurve)(E) *
    vignettingFactor(E, yz(theta, phi), jMax, kMax);
} // LowAccuracyEffectiveAreaDataServer::effectiveArea()


LowAccuracyEffectiveAreaDataServer::~LowAccuracyEffectiveAreaDataServer()
{
  unsigned jMax = _coefsCurve.xExtent(),
    kMax = _coefsCurve.yExtent();
  for (unsigned j=0; j<jMax; ++j)
    for (unsigned k=0; k<kMax; ++k)
      delete _coefsCurve[j][k];
} // LowAccuracyEffectiveAreaDataServer::~LowAccuracyEffectiveAreaDataServer()



MediumAccuracyEffectiveAreaDataServer::MediumAccuracyEffectiveAreaDataServer( 
               Proxy<EffectiveAreaCurve> onAxisEffectiveAreaCurve, 
               const DataSet *dataSet,
               CalAtomCollection<EffectiveAreaDataServer> *parent)
  : CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
    EffectiveAreaDataServer(parent),
    _onAxisEffectiveAreaCurve(onAxisEffectiveAreaCurve)
{

  // Read in the vignetting factors
  CcfConstituentTable ccftab(dataSet, "VIGNETTING");
  const Table *vignTable = ccftab.block();
  const ColumnData *cdE = vignTable->column("ENERGY")->data(),
  *cdVignFac = vignTable->column("VIGNETTING_FACTOR")->data();
  real32 *E = cdE->real32Data(),
    *vignFac = cdVignFac->real32Data();

  unsigned Ne = vignTable->rows(),
    Ntheta = cdVignFac->elements();
  _eGrid.resize(Ne);
  vector<double> thetaGrid(Ntheta);
  double deltaTheta = vignTable->attribute("D_THETA")->asReal() // [deg]
    * 3600.; // [deg->arcsec]
  for (unsigned n=0; n<Ntheta; ++n)
    thetaGrid[n] = DBL(n) * deltaTheta;
  thetaGrid.push_back(1.e4);	//	one data point at extreme angle to avoid
  //	extrapolation artefacts

  vector<double> index(Ne);
  for (unsigned ne=0; ne<Ne; ++ne) {
    index[ne] = ne;
    _eGrid[ne] = E[ne];
    vector<double> fac(Ntheta);
    for (unsigned nt=0; nt<Ntheta; ++nt)
      fac[nt] = vignFac[cdVignFac->position(ne, nt)];
    fac.push_back(0.);	//	vignetting 0 at theta=1.e4"

    _vignetting.push_back(new Slatec::CubicHermiteInterpolant<double>
			  (thetaGrid, fac));
  }
  _vignettingIndexLocator = new Slatec::CubicHermiteInterpolant<double>
    (_eGrid, index);

  delete cdE;
  delete cdVignFac;

  //    get the rectification values if any 

  //vector<double> rectmjd;
  if (dataSet->hasBlock("CORRAREA")){
     const Table *rectTab = dataSet->table("CORRAREA");
     //CalOalUtility::fillVector(rectTab, "MJD", rectmjd);
     CalOalUtility::fillVector(rectTab, "ENERGY", _areaCorrEnergyGrid);
     CalOalUtility::fillVector(rectTab, "FACTOR", _areaCorrCorrectionValues);
  }


#if 0
  unsigned N = eGrid.size();
  if (flatFields.size() != N)
    errstr<<"Vector with flatfields and energy grid vector have different "
      "sizes"<<Sas::error("SizeMismatch");

  //
  //	construct the interpolation maps
  //
  for (unsigned n=0; n<N; ++n) {
    const SImage &map = flatFields[n];
    int xExt = map.xExtent(),
      yExt = map.yExtent();
    vector<float> x(xExt),
      y(yExt),
      z(map.data(), map.data()+ xExt * yExt);
    for (int i=0; i<xExt; ++i)
      x[i] = DEG_TO_RAD*((i-INT(refPixXY)+1)*deltaXY + refValXY);
    for (int j=0; j<yExt; ++j)
      y[j] = DEG_TO_RAD*((j-INT(refPixXY)+1)*deltaXY + refValXY);

    _flatFields.push_back(
			  new Slatec::Bivariate2DInterpolantRectangular<float>(x, y, z));
  }
#endif
} // MediumAccuracyEffectiveAreaDataServer::MediumAccuracyEffectiveAreaDataSercer()


EffectiveAreaCurve MediumAccuracyEffectiveAreaDataServer::effectiveAreaCurve(
									     double theta,
									     double) const
{
  unsigned N = _eGrid.size();
  std::vector<double> vignetting(N);
  double abstheta = fabs(theta);
  for (unsigned n=0; n<N; ++n)
    vignetting[n] = (*_vignetting[n])(abstheta);

  Slatec::CubicHermiteInterpolant<double> vignFunc(_eGrid, vignetting),
    onAxisAe(*_onAxisEffectiveAreaCurve);

  return onAxisAe * vignFunc;
} // MediumAccuracyEffectiveAreaDataServer::effectiveAreaCurve()


inline unsigned MediumAccuracyEffectiveAreaDataServer::vignettingIndex(
								       double E) const
{
  int i = CalOalUtility::closestInt((*_vignettingIndexLocator)(E)),
    imax = _eGrid.size()-1;
  --i;

  if (i<0)
    return 0;
  if (i+3>imax)
    return UNT(imax-3);

  return UNT(i);
} // MediumAccuracyEffectiveAreaDataServer::vignettingIndex()


double MediumAccuracyEffectiveAreaDataServer::effectiveArea(
							    double E, double theta, double) const
{
  unsigned i = vignettingIndex(E);
  double ft = fabs(theta);
  double xp[] = { _eGrid[i], _eGrid[i+1], _eGrid[i+2], _eGrid[i+3] },
    yp[][1] = { {(*_vignetting[i])(ft)}, {(*_vignetting[i+1])(ft)},
		{(*_vignetting[i+2])(ft)}, {(*_vignetting[i+3])(ft)} },
      vig[0];

      Interpolator<1>::doPoly4(xp, yp, E, vig);

      SAS_DEBUG(cout<<xp[0]<<' '<<yp[0][0]<<endl;
		cout<<xp[1]<<' '<<yp[1][0]<<endl;
		cout<<xp[2]<<' '<<yp[2][0]<<endl;
		cout<<xp[3]<<' '<<yp[3][0]<<endl;
		cout<<"vignetting factor: "<<vig[0]<<endl;)

	return (*_onAxisEffectiveAreaCurve)(E) * vig[0];
} // MediumAccuracyEffectiveAreaDataServer::effectiveArea()


MediumAccuracyEffectiveAreaDataServer::~MediumAccuracyEffectiveAreaDataServer()
{
  unsigned N = _vignetting.size();
  for (unsigned n=0; n<N; ++n)
    delete _vignetting[n];
  delete _vignettingIndexLocator;
} // MediumAccuracyEffectiveAreaDataServer::~MediumAccuracyEffectiveAreaDataServer()

/*
 * applyAreaCorr - mutiplies an input effective area
 *                 curve with an artificial correction array
 */
EffectiveAreaCurve MediumAccuracyEffectiveAreaDataServer::applyAreaCorr(
                                 EffectiveAreaCurve eain) const
{

     EffectiveAreaCurve eaout;

     //	apply overall empirical correction curve
      unsigned N = _areaCorrEnergyGrid.size();
      if (N) {

        // Multiply the effarea curve with the area correction curve
         eaout = eain * Slatec::CubicHermiteInterpolant<double>(
                    _areaCorrEnergyGrid, _areaCorrCorrectionValues);
      }
      else
      {
    	   errstr<<"The present effective area correction CCF (XAREAEF)"
    		  <<" does not contain the CORRAREA structure - "
    		  <<"please check your CCF release"
    		  << Sas::error("MissingCORRAREA");
      }

     return eaout;
}

//
//	-----( EMOS/RGS EffectiveAreaDataServer )---------------------------------
//
EmosEffectiveAreaDataServer::EmosEffectiveAreaDataServer(
							 Proxy<EffectiveAreaDataServer> mds,
							 CalAtomCollection<EffectiveAreaDataServer> *parent)
  : CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
    EffectiveAreaDataServer(parent),
    _unvignetted(mds)
{
  GratingDataServer *gds = 0; gds = calServer.getAtom(gds);
  _gds = gds;
} // EmosEffectiveAreaDataServer::EmosEffectiveAreaDataServer()


EffectiveAreaCurve EmosEffectiveAreaDataServer::effectiveAreaCurveFull(
								       double theta,
								       double phi)
  const
{
  return _unvignetted->effectiveAreaCurve(theta, phi);
} // EmosEffectiveAreaDataServer::effectiveAreaCurveFull()


EffectiveAreaCurve EmosEffectiveAreaDataServer::effectiveAreaCurve(
								   double theta,
								   double phi) const
{
  InterpolatedCurve oef = _gds->obscurationEfficiency(theta, phi);
  EffectiveAreaCurve ae(_unvignetted->effectiveAreaCurve(theta, phi));
  ae *= (1. - oef);

  return ae;
} // EmosEffectiveAreaDataServer::effectiveAreaCurve()


double EmosEffectiveAreaDataServer::effectiveArea(double E,
						  double theta, double phi)
  const
{
  return _unvignetted->effectiveArea(E, theta, phi) *
    (1. - _gds->obscurationEfficiency(E, theta, phi));
} // EmosEffectiveAreaDataServer::effectiveArea()

EffectiveAreaCurve EmosEffectiveAreaDataServer::applyAreaCorr(
                                 EffectiveAreaCurve eain) const
{
  return _unvignetted->applyAreaCorr(eain);
}

RgsEffectiveAreaDataServer::RgsEffectiveAreaDataServer(
						       Proxy<EffectiveAreaDataServer> mds,
						       CalAtomCollection<EffectiveAreaDataServer> *parent)
  : CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
    EffectiveAreaDataServer(parent),
    _unvignetted(mds)
{

  //
  //	empirical Ae correction from table RGS?_QUANTUMEF.CCF:RGA_EFFAREACORR
  //
  CifEntry *qe = calServer.ccf()->entry(calServer.instrument(),
					Ccf::QuantumEf);
  DataSet *set = qe->open();
  if (set->hasBlock("RGA_EFFAREACORR")) {
    //	table is present - read contents
    Table *table = set->table("RGA_EFFAREACORR");
    std::vector<double> e, f;
    CalOalUtility::fillVector(table, "ENERGY", _empiricalCorrectionEgrid);
    CalOalUtility::fillVector(table, "FACTOR", _empiricalCorrectionValue);
  }
  dataSetServer->close(set);

  /*
   * AI 17/02/06: New implementation of the Effective Area Correction for RGS
   * The New EFFAREACORR CCF files have been created base on chebyshev polinomial
   * coorection. the chebyshev correction depends on a epoch. Therefore we
   * have to read all corrections that depends on a mjd and store its for a later
   * use.
   */

  // Get the RGS correction values CCF file
  CifEntry *e = calServer.ccf()->entry(calServer.instrument(),
				       Ccf::EffAreaCorr);
  DataSet *dataSet = e->open();

  // make sure we have a good CCF
  if (!dataSet->hasBlock("EFFAREACORR"))
    errstr<<"The present effective area correction CCF `"
	  <<dataSet->qualifiedName()
	  <<" does not contain the correct structure - "
	  <<"please check  your CCF release"
	  <<Sas::error("ImproperCcfConstituent");

  // Read the MJD and chebychev polynomial column data
  _hasEffAreaCorr = false;
  std::vector<double> mjd;
  std::vector<std::vector<double> > eGrid;
  std::vector<std::vector<double> > corrValues;

  if (dataSet->hasBlock("EFFAREACORR")){
	  const Table *chebTab = dataSet->table("EFFAREACORR");
	  CalOalUtility::fillVector(chebTab, "MJD", mjd);
	  CalOalUtility::fillMatrix(chebTab, "ENERGY", eGrid);
	  CalOalUtility::fillMatrix(chebTab, "FACTOR", corrValues);
	  _hasEffAreaCorr = true;
	  // Read the energy range keywords
	   double emin = chebTab->attribute("EMIN")->asReal();
	   double emax = chebTab->attribute("EMAX")->asReal();
	   _erange = DPair(emin,emax);
	   // Calculate the effective area curves (and rectification curves)
	   // for this observation epoch
	   errstr << "Epoch Interpolation for Effective Area Correction..."
	 	 << Sas::message(Msg::App, Msg::Verbose);
	   vector<vector<double> > effVals = epochInterpolation(mjd,eGrid,corrValues);
	   _realisticEnergyGrid = effVals[0];
	   _realisticCorrectionValues = effVals[1];
  }

  // Read in the rectification array
  _hasRect = false;
  vector<double> rectmjd;
  vector<vector<double> > recteGrid;
  vector<vector<double> > rectcorrValues;
  if (dataSet->hasBlock("RECTIFICATION")){
     const Table *rectTab = dataSet->table("RECTIFICATION");
     CalOalUtility::fillVector(rectTab, "MJD", rectmjd);
     CalOalUtility::fillMatrix(rectTab, "ENERGY", recteGrid);
     CalOalUtility::fillMatrix(rectTab, "FACTOR", rectcorrValues);
     _hasRect = true;
     errstr << "Epoch Interpolation for Rectification Correction..."
         << Sas::message(Msg::App, Msg::Verbose);
     vector<vector<double> > rectVals = epochInterpolation(rectmjd,recteGrid,rectcorrValues);
     _rectifyEnergyGrid = rectVals[0];
     _rectifyCorrectionValues = rectVals[1];

  }

  // Read in the Effective area correction array
  _hasRGSEffAreaCorr = false;
  std::vector<double> effAreaCorrmjd;
  std::vector<std::vector<double> > effAreaRecteGrid;
  std::vector<std::vector<double> > effAreaRectCorrValues;

  if (dataSet->hasBlock("AREACORR_1")){
     const Table *rectTab = dataSet->table("AREACORR_1");
     CalOalUtility::fillVector(rectTab, "MJD", effAreaCorrmjd);
     CalOalUtility::fillMatrix(rectTab, "ENERGY", effAreaRecteGrid);
     CalOalUtility::fillMatrix(rectTab, "FACTOR", effAreaRectCorrValues);
     _hasRGSEffAreaCorr = true;
     errstr << "Epoch Interpolation for RGS Effective Area Correction..."
         << Sas::message(Msg::App, Msg::Verbose);
     vector<vector<double> > effAreaRectVals = epochInterpolation(effAreaCorrmjd,effAreaRecteGrid,effAreaRectCorrValues);
     _effAreaCorrEnergyGrid_o1 = effAreaRectVals[0];
     _effAreaCorrValues_o1 = effAreaRectVals[1];

  }

  if (dataSet->hasBlock("AREACORR_2")){
     const Table *rectTab = dataSet->table("AREACORR_2");
     CalOalUtility::fillVector(rectTab, "MJD", effAreaCorrmjd);
     CalOalUtility::fillMatrix(rectTab, "ENERGY", effAreaRecteGrid);
     CalOalUtility::fillMatrix(rectTab, "FACTOR", effAreaRectCorrValues);
     _hasRGSEffAreaCorr = true;
     errstr << "Epoch Interpolation for RGS Effective Area Correction..."
         << Sas::message(Msg::App, Msg::Verbose);
     vector<vector<double> > effAreaRectVals = epochInterpolation(effAreaCorrmjd,effAreaRecteGrid,effAreaRectCorrValues);
     _effAreaCorrEnergyGrid_o2 = effAreaRectVals[0];
     _effAreaCorrValues_o2 = effAreaRectVals[1];

  }

  // Close the CCF
  dataSetServer->close(dataSet);

} // RgsEffectiveAreaDataServer::RgsEffectiveAreaDataServer()


EffectiveAreaCurve RgsEffectiveAreaDataServer::effectiveAreaCurveFull(
								      double theta,
								      double phi)
  const
{
  return _unvignetted->effectiveAreaCurve(theta, phi);
} // RgsEffectiveAreaDataServer::effectiveAreaCurveFull()


EffectiveAreaCurve RgsEffectiveAreaDataServer::effectiveAreaCurve(
								  int spectralOrder,
								  double theta,
								  double phi) const
{
  EffectiveAreaCurve ae(intrinsicEffectiveAreaCurve(spectralOrder,
						    theta, phi));

  GratingDataServer *gds = 0; gds = calServer.getAtom(gds);
  ae *= gds->selfVignettingEfficiency(ae.abscissa(), theta, phi,
				      spectralOrder);

  //
  //	[2]: missing are last two factors:
  //		CAL_rgsCrossPsf x f_xdspsel
  //		CAL_rgsLSF x f_prdersel
  //

  return ae;
} // RgsEffectiveAreaDataServer::effectiveAreaCurve()


EffectiveAreaCurve RgsEffectiveAreaDataServer::intrinsicEffectiveAreaCurve(
									   int spectralOrder,
									   double theta,
									   double phi) const
{
  EffectiveAreaCurve ae(_unvignetted->effectiveAreaCurve(theta, phi));

  GratingDataServer *gds = 0; gds = calServer.getAtom(gds);
  ae *= gds->interceptEfficiency(theta, phi);
  ae *= gds->reflectionEfficiency(theta, phi, spectralOrder);

  //	apply overall empirical correction curve
  unsigned N = _empiricalCorrectionEgrid.size();
  if (N) {
    //	modify Ae curve by interpolated empirical correction data
    return ae *
      Slatec::CubicHermiteInterpolant<double>(_empiricalCorrectionEgrid,
					      _empiricalCorrectionValue);
    //
    //	old scheme
    //		vector<double> vadd(N);
    //		for (unsigned n=0; n<N; ++n)
    //			vadd[n] = ae(_empiricalCorrectionEgrid[n]) *
    //												_empiricalCorrectionValue[n];
    //		ae.supplement(_empiricalCorrectionEgrid, vadd);
  }

  return ae;
} // RgsEffectiveAreaDataServer::intrinsicEffectiveAreaCurve()


/*
 * AI:This method uses the new Effective Area Corr based on the Chebyshev
 * polinomials.
 */

EffectiveAreaCurve RgsEffectiveAreaDataServer::realisticEffectiveAreaCurve(
									   int spectralOrder,
									   double theta,
									   double phi) const
{
  EffectiveAreaCurve ae(_unvignetted->effectiveAreaCurve(theta, phi));

  GratingDataServer *gds = 0; gds = calServer.getAtom(gds);
  ae *= gds->interceptEfficiency(theta, phi);
  ae *= gds->reflectionEfficiency(theta, phi, spectralOrder);

  //	apply overall empirical correction curve
  unsigned N = _realisticEnergyGrid.size();
  if (N) {
    //	modify Ae curve by interpolated empirical correction data
    return ae *
      Slatec::CubicHermiteInterpolant<double>(_realisticEnergyGrid,
					      _realisticCorrectionValues);
  }

  return ae;
} // RgsEffectiveAreaDataServer::realisticEffectiveAreaCurve()

/*
 * EffectiveAreaRectification - mutiplies an input effective area
 *                              curve with a rectification array
 */
EffectiveAreaCurve RgsEffectiveAreaDataServer::effectiveAreaRectification(
                                 EffectiveAreaCurve eain) const
{

     EffectiveAreaCurve eaout;
     // apply overall empirical correction curve
      unsigned N = _rectifyEnergyGrid.size();
      if (N) {
     // Multiply the effarea curve with the rectification curve
      eaout = eain * Slatec::CubicHermiteInterpolant<double>(_rectifyEnergyGrid,
                  _rectifyCorrectionValues);
      }
      else
      {
           errstr<<"The present effective area correction CCF (EFFAREACORR)"
                  <<" does not contain the RECTIFICATION structure - "
                  <<"please check your CCF release"
                  << Sas::error("MissingRECTIFICATION");
      }

     return eaout;
}


std::vector<double > RgsEffectiveAreaDataServer::effectiveAreaCorrection(int order ) const
{

	if (order > 2)
	{
        errstr<<"The present effective area correction CCF (EFFAREACORR)"
               <<" does not contain corrector factors for spectral order " << order << ". "
               << Sas::error("MissingEFFECTIVEAREACORRECTION");
	}

	if (order == 1)
		return  _effAreaCorrValues_o1;
	else if (order == 2 )
		return  _effAreaCorrValues_o2;


}



std::vector<double > RgsEffectiveAreaDataServer::effectiveAreaCorrectionEner(int order ) const
{

	if (order > 2)
	{
        errstr<<"The present effective area correction CCF (EFFAREACORR)"
               <<" does not contain corrector factors for spectral order " << order << ". "
               << Sas::error("MissingEFFECTIVEAREACORRECTION");
	}

	if (order == 1)
		return  _effAreaCorrEnergyGrid_o1;
	else if (order == 2 )
		return  _effAreaCorrEnergyGrid_o2;

}

std::vector<std::vector<double> > RgsEffectiveAreaDataServer::epochInterpolation(vector<double> mjd,
		std::vector<std::vector<double> > eGrid, vector<vector<double> > corrValues )
{

  STime obsStartTime = calServer.state()->exposureStartTime()->get();
  errstr << "Current Observation Start Time "
	 << obsStartTime.modifiedJulianDay().full()
	 << Sas::message(Msg::App, Msg::Verbose);

  double currentObsDate = obsStartTime.modifiedJulianDay().full();

  errstr << "Available correction date... "
	 << Sas::message(Msg::App, Msg::Verbose);
  for (unsigned int i = 0; i < mjd.size() ; i++)
    errstr << " MJD: " << mjd[i]
	   << Sas::message(Msg::App, Msg::Verbose);


  /*
   * Swap the existing eGrid and corrValues vectors
   */
  vector<vector<double> > energy_rotated;
  vector<double> tmpEnergy;

  int counter = 0;

  while (counter <= (int) corrValues[0].size())
    {
      for (unsigned int i = 0 ; i < eGrid.size() ; i++)
    	  tmpEnergy.push_back(eGrid[i][counter]);

      energy_rotated.push_back(tmpEnergy);
      tmpEnergy.clear();
      counter++;
    }

  /*
   * Swap vector of columns to a vector of rows.
   */
  eGrid = energy_rotated;

  vector<vector<double> > corrValues_rotated;
  vector<double> tmpCorr;

  counter = 0;
  while (counter <= (int) corrValues[0].size())
    {
      for (unsigned int i = 0 ; i < corrValues.size() ; i++)
    	  tmpCorr.push_back(corrValues[i][counter]);

      corrValues_rotated.push_back(tmpCorr);
      tmpCorr.clear();
      counter++;
    }

  /*
   * Swap vector of columns to a vector of rows.
   */
  corrValues = corrValues_rotated;

  /*
   * End of Swaping
   */

  // Calculate the epoch interval needed for current Obs date

  unsigned int epochCounter = 0;

  while( (currentObsDate >= mjd[epochCounter]) &&
	 (epochCounter != mjd.size()) )
    epochCounter++;

  /*
   * output arrays
   */
  vector<double> outputEnerGrid,outputCorrValues;
  vector<vector<double> > ouptutValues;

  if (epochCounter == 0)
    {
      errstr << "Epoch interpolation not needed. "
	     << "Using correction for date: "
	     << mjd[epochCounter]
	     << Sas::message(Msg::App, Msg::Verbose);

      for (unsigned int i = 0 ; i < eGrid[0].size() ; i++)
      {
    	  outputCorrValues.push_back(corrValues[epochCounter][i]);
    	  outputEnerGrid.push_back(eGrid[0][i]*1000.0);
      }

    }
  else if (epochCounter == mjd.size() )
    {
      errstr << "Epoch interpolation not needed. "
	     << "Using correction for date: "
	     << mjd[epochCounter-1]
	     << Sas::message(Msg::App, Msg::Verbose);

      for (unsigned int i = 0 ; i < eGrid[0].size() ; i++)
      {
    	  outputCorrValues.push_back(corrValues[epochCounter-1][i]);
    	  outputEnerGrid.push_back(eGrid[0][i]*1000.0);
      }
    }
  else
    {
      errstr << "Epoch Interval: " << mjd[epochCounter-1]
	     << " -- " << mjd[epochCounter]
	     << Sas::message(Msg::App, Msg::Verbose);

      /*
       * Linear Interpolation Weight Factor
       */

      errstr << "Calculating Weight Factor... "
	     << Sas::message(Msg::App, Msg::Verbose);

      double weight;
      weight = 1.0 - ( mjd[epochCounter] - currentObsDate ) /
		  (mjd[epochCounter] - mjd[epochCounter-1] );

      errstr << "Weight Factor: " << weight
	     << Sas::message(Msg::App, Msg::Verbose);

      double tmp;
      for (unsigned int i = 0 ; i < eGrid[0].size() ; i++)
      {
    	  tmp = (1.0 - weight)*corrValues[epochCounter-1][i] +
    	  weight*corrValues[epochCounter][i];
    	  outputCorrValues.push_back(tmp);
    	  outputEnerGrid.push_back(eGrid[0][i]*1000.0);
      }

      errstr << "Interpolated Coefficients for date: " << currentObsDate
	     << Sas::message(Msg::App, Msg::Verbose);
      errstr << "Energy Grid -- Correction Values "
	     << Sas::message(Msg::App, Msg::Verbose);
      for (unsigned int i = 0 ; i < outputEnerGrid.size() ; i++)
    	  errstr << outputEnerGrid[i] << "  " << outputCorrValues[i]
	       << Sas::message(Msg::App, Msg::Verbose);
	}

  ouptutValues.push_back(outputEnerGrid);
  ouptutValues.push_back(outputCorrValues);

  return ouptutValues;

}

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<EffectiveAreaDataServer>::CalAtomCollection(
							      const XmmInstrument *instrument,
							      const Ccf *ccf,
							      Xmm::Filter filter,
							      AccuracyLevel accuracy)
{
  _filter		= filter;
  _accuracy	= accuracy;
  _nCCDs		= instrument->nCcds();
  _nNodes		= instrument->nNodes();
  _nEntries	= 1;
  EffectiveAreaDataServer *null = 0;
  _collection = new vector<EffectiveAreaDataServer *>(_nEntries, null);

  CifEntry *e = ccf->entry(instrument, Ccf::XAreaEf);
  DataSet *set = e->open();

  //
  //	read on-axis effective area data
  //
  CcfConstituentTable ccftab(set, "ONAXISXAREAEF");
  const Table *table = ccftab.block();
  unsigned N = table->rows();
  vector<double> Evec(N),
    onAxisAe(N);
  ColumnData *cdE = table->column("ENERGY")->data(),
    *cdAe = table->column("AREA")->data();
  real32 *E = cdE->real32Data(),
    *area = cdAe->real32Data();
  for (unsigned n=0; n<N; ++n) {
    Evec[n] = E[n];
    onAxisAe[n] = area[n];
  }
  delete cdE;
  delete cdAe;

  EffectiveAreaDataServer *mds;
  Proxy<EffectiveAreaCurve> ae = new EffectiveAreaCurve(Evec, onAxisAe);

  AccuracyLevel acc = MEDIUM;	//	for the moment disable low-accuracy
  //	EffectiveArea data server; numerical
  //	problems with polynomial fit from IDL
  //	= _accuracy
  switch (acc) {
  case LOW: {
    CcfConstituentTable ccftab(set, "FIT_PARAM");
    const Table *table = ccftab.block();
    const ColumnData *cdE = table->column("ENERGY")->data(),
      *cdCoef = table->column("COEF_MAPS")->data();
    real32 *E = cdE->real32Data();
    real64 *coef = cdCoef->real64Data();

    unsigned N = table->rows();
    vector<double> eGrid(N);
    vector<DImage> coefMaps;
    vector<unsigned long> exts = cdCoef->size();
    for (unsigned n=0; n<N; ++n) {
      eGrid[n] = E[n];
      coefMaps.push_back(DImage(exts[0], exts[1]));
      double *map = coefMaps[n].data();
      for (unsigned i=0; i<exts[0]*exts[1]; ++i)
	map[i] = coef[cdCoef->position(n, i)];
    }
    delete cdE;
    delete cdCoef;

    mds = new LowAccuracyEffectiveAreaDataServer(
						 ae,
						 eGrid,
						 coefMaps,
						 table->attribute("DELTA_F")->asReal(),
						 table->attribute("DELTA_Y")->asReal(),
						 table->attribute("DELTA_Z")->asReal(),
						 this);

    break;
  }
  case HIGH:
    errstr<<"High-accuracy version of EffectiveArea effective area is "
      "not available - reverting to normal accuracy"
	  <<Sas::warning("NotAvailable");
  case MEDIUM: {
    mds = new MediumAccuracyEffectiveAreaDataServer(ae,
						    set,
						    this);
    break;
  }
  default:
    errstr<<"Unknown accuracy"<<Sas::error("ThisShouldNotHappen");
    break;
  }

  dataSetServer->close(set);

  //
  //	now instantiate right EffectiveArea data server
  //
  switch (instrument->id()) {
  case Xmm::EMOS1:
  case Xmm::EMOS2:
    (*_collection)[0] =
      new EmosEffectiveAreaDataServer(
				      Proxy<EffectiveAreaDataServer>(mds), this);
    break;
  case Xmm::RGS1:
  case Xmm::RGS2:
    (*_collection)[0] =
      new RgsEffectiveAreaDataServer(
				     Proxy<EffectiveAreaDataServer>(mds), this);
    break;
  default:
    //	for PN the Ae is the full one
    (*_collection)[0] = mds;
  }
} // CalAtomCollection<EffectiveAreaDataServer>::CalAtomCollection()


/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
EffectiveAreaDataServer *CalAtomCollection<EffectiveAreaDataServer>::select(
									    Ccd,
									    Node,
									    unsigned) const
{
  return (*_collection)[0];
} // CalAtomCollection<EffectiveAreaDataServer>::select()
