// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EnergyCorrector.cc --- implementation of EnergyCorrector class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-05-25: 1.0.0: - first version (extracted from
//								 CalAtomsCommon.cc)
//		UL, 1999-06-08: 1.1.0: - minor bug fixes
//		UL, 1999-06-15: 1.2.0: - added implementation of EpnEnergyCorrector
//							     class methods
//      UL, 1999-08-13: 1.3.0: - added piTokeV()
//		UL, 1999-09-03: 1.4.0: - changed ser->grade in
//								   RgsEnergyCorrector::offsetCorrect and
//								 updated implementation
//		UL, 1999-11-24: 1.5.0: - brought in line with calpnalgo: pass
//								 additional parameter vector to gain init
//								 routine
//		UL, 2000-01-05: 1.6.0: - removed check that cross-tests pattern
//								 IDs in energy combinator coefficient
//								 columns against those in pattern lib
//								 (fixes SPR SPR 1052 : CAL_gainCorrect does
//								  not work in Timing mode for EMOS)
//		UL, 2000-02-01: 1.7.0: - made EPN-specific gainCorrect() in generic
//								 corrector use generic gainCorrect()
//		UL, 2000-02-29: 1.8.0: - brought RGS offset correction in line with
//								 [2]
//		UL, 2000-03-01: 1.9.0: - RgsEnergyCorrector::offsetCorrect: bug
//								 fix + more sensible handling of illegal
//								 event shapes
//		UL, 2000-04-04: 1.10.0:- E_RAND_MIN/E_RAND_MAX -> RAND_MIN/RAND_MAX
//		UL, 2000-04-13: 1.11.0:- check error code from PN gain correction
//		UL, 2000-05-10: 1.12.0:- overhaul in the process of implementing new
//								 MOS gain correction
//		UL, 2000-05-22: 1.13.0:- fixed problem with generic/EMOS-specific
//								 gain correction routine
//							   - fall back upon CCF if given RGS offset
//								 override table is empty
//		UL, 2000-07-31: 1.14.0:- EPN gain correction takes now real-valued
//								 input amplitude
//		UL, 2000-11-08: 1.15.0:- EmosEnergyCorrector::gainCorrect() in modes
//								 FastUncompressed/FastCompressed for CCD 1
//								 takes additional multiplicative factor
//								 (FAST_G in CCF::AduConv) into accoun
//		UL, 2000-11-14: 1.15.1:- bug fix in
//									RgsEnergyCorrector::RgsEnergyCorrector
//								 [fixes SPR 1834: Task aborts with Bus Error
//		UL, 2000-11-21: 1.15.2:- PN gain correction: better message in case
//								 of calpnalgo failure
//		UL, 2001-04-03: 1.16.0:- several changes related to ALGOID=1
//								 RGS offset correction; offset values
//								 are now time dependent:
//									offset = offsetConst + t*offsetSlope
//								 t = obsStartDate - refTime
//								 refTime comes from ADUCONV
//		UL, 2001-10-30: 1.17.0:- added implementation of new RGS HTR offset
//								 correction
//		UL, 2001-11-21: 1.18.0:- fixed problem in HTR-specific offset
//								 correction: was testing against wrong
//								 camera mode
//		UL, 2002-07-17: 1.19.0:- added three new pn-specific methods (see
//								 EnergyCorrector.h)
//		UL, 2002-11-19: 1.20.0:- implemented ALGOID=1 of EPN long-term CTE
//								 model (time-dependence modelled
//								 as 3. order polynomial; parameters mode
//								 dependent); goes with structural change
//								 in EPN_CTI CCF; all changes fully backward
//								 compatible
//		UL, 2003-02-12: 1.20.1:- fixed bug in EnergyCorrector::offsetCorrect();
//								 only surfaced with gcc-3.2
//		RDS,2004-08-04: 1.21.0:- added gain correction for PN Eff and FF mode
//
//		RDS,2008-04-23: 1.22.0:- added Rate-dependent CTI correction
//                                       for PN Fast modes
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd et al.
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/EnergyCorrector.cc,v 1.68 2014/03/19 15:50:24 rsaxton Exp $
//

#include <iomanip>
#include <algorithm>
#include <cmath>

#include "errstr.h"
#include "common.h"
#include "EnergyCorrector.h"
#include "CalServer.h"
#include "CalOalUtility.h"
#include "BadPixelDataServer.h"	//	INVALID_BAD_PIXEL_TABLE_CODE
#include "CalPnAlgo.h"
#include "slatec++.h"

using namespace std;

//
//	---( EnergyCorrector )---------------------------------------------------
//
/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<EnergyCorrector>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	EnergyCorrector *null = 0;
	_collection = new vector<EnergyCorrector *>(_nEntries, null);

	//
	//	read the principal data from the CCF
	//
	CifEntry *e = ccf->entry(instrument, Ccf::ADUConv);
	DataSet *dataSet = e->open();
	CcfConstituentTable aduConvTab(dataSet,
								   e->queryBlockName("OFFSET_GAIN"),
								   1);	//	we know about ALGOID: 0+1

	//
	//	instantiate high-accuracy/instrument-specific gain corrector
	//
	switch (instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2: {
			CcfConstituentTable ccftab(dataSet, "ENERGY_COMBINE");
			const Table *eCombine = ccftab.block();
			for (unsigned n=0; n<_nEntries; ++n)
				(*_collection)[n] =
					new EmosEnergyCorrector(n+1, aduConvTab, eCombine, this);
			}
			break;
		case Xmm::EPN: {
			CcfConstituentTable ccftab1(dataSet, "GAIN-HIGH"),
								ccftab2(dataSet, "REEMISSION");
			const Table *gainHigh = ccftab1.block(),
					    *reEmission = ccftab2.block();
			for (unsigned n=0; n<_nEntries; ++n)
				(*_collection)[n] =
					new EpnEnergyCorrector(n+1, aduConvTab, gainHigh,
										   reEmission, ccf, this);
			}
			break;
		case Xmm::RGS1:
		case Xmm::RGS2: {
			CifEntry *e = ccf->entry(instrument, Ccf::DarkFrame);
			DataSet *darkSet = e->open(),
					*diagSet = 0;
			CcfConstituentTable ccftab(darkSet, "DARKFRAME");
			const Table *darkFrame = ccftab.block(),
						*offsetOverride = 0;
			if (ccf->hasEntry(instrument, Ccf::DiagnosticOffset)) {
				CifEntry *e = ccf->entry(instrument, Ccf::DiagnosticOffset);
				diagSet = e->open();
				const Table *offsetTab = diagSet->table("RGSOFFSETS");
				if (offsetTab->rows())
					offsetOverride = offsetTab;
				else
					errstr<<"Specified offset override table `"
						  <<offsetTab->qualifiedName()<<"' is empty (0 rows) "
							"- offset correction will fall back upon numbers "
							"in the CCF"
						  <<Sas::warning("EmptyOffsetOverrideTable");
			}
			for (unsigned n=0; n<_nEntries; ++n) {
				(*_collection)[n] =
					new RgsEnergyCorrector(n+1, aduConvTab, darkFrame,
										   offsetOverride, this);
			}
			if (diagSet)
				dataSetServer->close(diagSet);
			dataSetServer->close(darkSet);
			}
			break;
		default:
			break;
	}

	dataSetServer->close(dataSet);
} // CalAtomCollection<EnergyCorrector>::CalAtomCollection()

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
EnergyCorrector *CalAtomCollection<EnergyCorrector>::select(
					Ccd ccd,
					Node node,
					unsigned) const
{
	_ccd = ccd;
	_node = node;
	unsigned n = ccd - 1;

	return (*_collection)[n];
} // CalAtomCollection<EnergyCorrector>::select


EnergyCorrector::EnergyCorrector(Ccd ccd,
								 const CcfConstituentTable &aduConvTab,
								 CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent)
{
	const Table *aduConv = aduConvTab.block();
	_bptCode = CalOalUtility::optionalAttribute(aduConv, "BPT_CODE",
						 BadPixelDataServer::INVALID_BAD_PIXEL_TABLE_CODE);

	//
    //  read PI->eV conversion values
    //
	_eVperPi = aduConv->parent()->attribute("KEV_P_PI")->asReal()*1000.;
	_eV0	 = aduConv->parent()->attribute("KEV0")->asReal()*1000.;

	ColumnData *cdccdId	= aduConv->column("CCD_ID")->data(),
			   *cdgain	= aduConv->column("GAIN")->data(),
			   *cdoffset0 = 0, *cdoffset1 = 0;

	_offsetRefTime = Xmm::missionReferenceTime;
	if (aduConvTab.algoid() == 0)
		cdoffset0 = aduConv->column("OFFSET")->data();
	else {
		cdoffset0 = aduConv->column("OFFSET_CONST")->data();
		cdoffset1 = aduConv->column("OFFSET_SLOPE")->data();
		_offsetRefTime = STime(aduConv->attribute("REFTIME")->asString());
	}

	int8	*ccdId		= cdccdId->int8Data();
	real32	*gain		= cdgain->real32Data(),
			*offset0	= cdoffset0->real32Data(),
			*offset1	= cdoffset1 ? cdoffset1->real32Data() : 0;

	for (unsigned n=0; n<aduConv->rows(); ++n) {
		unsigned nn = cdgain->elements();
		if (ccdId[n] <= 0 || ccdId[n] > parent->ccds() ||
			nn > parent->nodes())
			errstr<<"Illegal chip and/or node number in row #"
				  <<n<<" of table `"<<aduConv->qualifiedName()<<'\''
			  <<Sas::error("IllegalChipNodeId");

		if (ccdId[n] == ccd) {
			for (unsigned j=0; j<nn; ++j) {
				unsigned p = cdgain->position(n, j);
				//
				//	check for illegal gain values
				//
				if (gain[p] <= 0.)
					errstr<<"gain for CCD #"<<ccdId[n]<<"/node #"<<j
						  <<" is less than or equal to zero"
						  <<Sas::error("IllegalGainValue");

				_offsetConst.push_back(offset0[p]);
				_offsetSlope.push_back(offset1 ? offset1[p] : 0.);
				_gain.push_back(gain[p]);
			}
			break;
		}
	}

	//
	//	check that we have read enough data
	//
	if (_gain.size() != parent->nodes() ||
		_offsetConst.size() != parent->nodes())
		errstr<<"Expected to read "<<parent->nodes()<<" values for the "
				"offset/gain from table `"
			  <<aduConv->qualifiedName()<<"' but actually found: "
			  <<_offsetConst.size()<<"/"<<_gain.size()
			  <<Sas::error("wrongNumberOfElements");

	//	tidy up
	delete cdccdId;
	delete cdgain;
	delete cdoffset0;
	delete cdoffset1;
} // EnergyCorrector::EnergyCorrector()


void EnergyCorrector::dump(ostream &os)
{
	os<<"EnergyCorrector{\n"
	  <<"offset: ";
	for (unsigned n=0; n<_offsetConst.size(); ++n)
		os<<_offsetConst[n]<<'/'<<_offsetSlope[n]<<' ';
	os<<"gain  : ";
	for (unsigned n=0; n<_gain.size(); ++n)
		os<<_gain[n]<<" ";

	os<<"}";
} // EnergyCorrector::dump()


float EnergyCorrector::offsetValue(Node node) const
{
	unsigned n = UNT(node);
	if (n>=_offsetConst.size())
		errstr<<"Unknown readout node "<<INT(node)
			  <<Sas::error("UnknownReadoutNode");

	double deltaT = calServer.state()->exposureStartTime()->get() -
						_offsetRefTime;

	return _offsetConst[n] + deltaT * _offsetSlope[n];
} // EnergyCorrector::offsetValue()


CalReal32Vector &EnergyCorrector::offsetCorrect(const CalReal64Vector &,
											  const CalInt8Vector &nodes,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	errstr<<"Detailed time dependence of offset correction currently not "
			"implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	unsigned nNodes = _offsetSlope.size();
	float offsetVec[nNodes];
	for (unsigned n=0; n<nNodes; ++n)
		offsetVec[n] = offsetValue(Xmm::Node(n));

	float offset;
	for (unsigned n=0; n<eIn.size(); ++n) {
		if (UNT(nodes[n]) < nNodes)
			offset = offsetVec[nodes[n]];
		else {
			errstr<<"No offset value for node "<<INT(nodes[n])
				  <<" event #"<<n<<" - no correction applied"
				  <<Sas::warning("NoSuchValue");
			offset = 0.f;
		}

		eOut[n] = eIn[n] - offset;
	}

	return eOut;
} // EnergyCorrector::offsetCorrect()


CalReal32Vector &EnergyCorrector::offsetCorrect(const CalReal64Vector &time,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	unsigned N=eIn.size();
	CalInt8Vector nodes(0, N);
	int nodeIndex = parent()->node();
	for (unsigned n=0; n<N; nodes[n++] = nodeIndex) ;

	offsetCorrect(time, nodes, eIn, eOut);

	return eOut;
} // EnergyCorrector::offsetCorrect()


CalReal32Vector &EnergyCorrector::offsetCorrect(const CalReal64Vector &,
											    const CalInt16Vector &,
												const CalInt8Vector &,
												const CalInt16Vector &,
											    CalReal32Vector &eOut) const
{
	errstr<<"Generic energy corrector cannot perform RGS HTR-specific "
			"offset correction"<<Sas::error("UnsupportedOperation");

	return eOut;
} // EnergyCorrector::offsetCorrect()


CalReal32Vector &EnergyCorrector::gainCorrect(const CalReal64Vector &time,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	unsigned N = eIn.size();
	CalInt8Vector nodes(0, N);
	int nodeIndex = parent()->node();
	for (unsigned n=0; n<N; nodes[n++] = nodeIndex) ;

	gainCorrect(time, nodes, eIn, eOut);

	return eOut;
} // EnergyCorrector::gainCorrect()



CalReal32Vector &EnergyCorrector::gainCorrect(const CalReal64Vector &,
											  const CalInt8Vector &nodes,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	errstr<<"Time dependence of gain correction not implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	char nNodes = _gain.size();
	float gain;

	for (unsigned n=0; n<eIn.size(); ++n) {
		if (nodes[n] < nNodes)
			gain = _gain[nodes[n]];
		else {
			errstr<<"No gain value for node "<<INT(nodes[n])
				  <<" event #"<<n<<" - no correction applied"
				  <<Sas::warning("NoSuchValue");
			gain = 1.f;
		}

		eOut[n] = eIn[n];
		if (gain != 0.)
			eOut[n] /= gain;
		else
			errstr<<"Gain is 0 for node "<<INT(nodes[n])<<" on event number "
				  <<n<<" - energy unmodified"
				  <<Sas::warning("ZeroGain");
	}

	return eOut;
} // EnergyCorrector::gainCorrect()


CalReal32Vector &EnergyCorrector::piToeV(const CalReal32Vector &eIn,
										 CalReal32Vector &eOut) const
{
	for (unsigned n=0; n<eIn.size(); ++n)
		// eOut[n] = _eV0 + eIn[n] * _eVperPi;

        /* In practise the conversion from PI to eV is not linear
           and the current implementation is wrong by a factor of 5.
           Rather than change the header values in many CCF files for
           simplicity we here hardcode the conversion of 1eV per channel.
           This WILL be documented */

        eOut[n] = eIn[n] * 1.0;

	return eOut;
} // EnergyCorrector::piToeV()

CalReal32Vector &EnergyCorrector::eVToPi(const CalReal32Vector &eIn,
                                         CalReal32Vector &eOut) const
{
    for (unsigned n=0; n<eIn.size(); ++n)
        eOut[n] = eIn[n] * 1.0;

    return eOut;
} // EnergyCorrector::eVToPi()

//
//	--------------------------( EMOS )--------------------------------------
//
vector<float> EmosEnergyCorrector::_e1Coef,
			  EmosEnergyCorrector::_e2Coef,
			  EmosEnergyCorrector::_e3Coef,
			  EmosEnergyCorrector::_e4Coef;
bool		  EmosEnergyCorrector::_coefRead = false;


EmosEnergyCorrector::EmosEnergyCorrector(Ccd ccd,
										 const CcfConstituentTable &aduConvTab,
										 const Table *eCombine,
								 CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent),
	  EnergyCorrector(ccd, aduConvTab, parent)
{
	const Table *aduConv = aduConvTab.block();

	//
	//	read offset/gain coefficients
	//
	ColumnData *cdCcdId		= aduConv->column("CCD_ID")->data(),
			   *cdOffset0	= aduConv->column("OFFSET")->data(),
			   *cdOffset1	= aduConv->column("OFFSET1")->data(),
			   *cdOffset2	= aduConv->column("OFFSET2")->data(),
			   *cdOffset3	= aduConv->column("OFFSET3")->data(),
			   *cdGain0		= aduConv->column("GAIN")->data(),
			   *cdGain1		= aduConv->column("GAIN1")->data(),
			   *cdGain2		= aduConv->column("GAIN2")->data(),
			   *cdGain3		= aduConv->column("GAIN3")->data(),
			   *cdPoly0		= aduConv->column("POLY_TERM")->data(),
			   *cdPoly1		= aduConv->column("POLY_TERM1")->data(),
			   *cdPoly2		= aduConv->column("POLY_TERM2")->data(),
			   *cdPoly3		= aduConv->column("POLY_TERM3")->data();

	int8	*ccdId = cdCcdId->int8Data();
	float	*offset0 = cdOffset0->real32Data(),
			*offset1 = cdOffset1->real32Data(),
			*offset2 = cdOffset2->real32Data(),
			*offset3 = cdOffset3->real32Data(),
			*gain0 = cdGain0->real32Data(),
			*gain1 = cdGain1->real32Data(),
			*gain2 = cdGain2->real32Data(),
			*gain3 = cdGain3->real32Data(),
			*poly0 = cdPoly0->real32Data(),
			*poly1 = cdPoly1->real32Data(),
			*poly2 = cdPoly2->real32Data(),
			*poly3 = cdPoly3->real32Data();

	unsigned nodes = parent->nodes();
	if (cdOffset0->elements() != nodes || cdOffset1->elements() != nodes ||
		cdOffset2->elements() != nodes || cdOffset3->elements() != nodes ||
		cdGain0->elements() != nodes || cdGain1->elements() != nodes ||
		cdGain2->elements() != nodes || cdGain3->elements() != nodes ||
		cdPoly0->elements() != nodes || cdPoly1->elements() != nodes ||
		cdPoly2->elements() != nodes || cdPoly3->elements() != nodes)
		errstr<<"Size of vector column in `"<<aduConv->qualifiedName()
			  <<"' does not coincide with number of nodes ("<<nodes<<")"
			  <<Sas::error("CcfDataIntegrityError");

	unsigned n;
	for (n=0; n<aduConv->rows(); ++n)
		if (ccdId[n] == ccd) {
			for (unsigned j=0; j<nodes; ++j) {
				unsigned p = cdOffset0->position(n, j);

				_offsetGain[0][0][j] = offset0[p];
				_offsetGain[0][1][j] = offset1[p];
				_offsetGain[0][2][j] = offset2[p];
				_offsetGain[0][3][j] = offset3[p];

				_offsetGain[1][0][j] = gain0[p];
				_offsetGain[1][1][j] = gain1[p];
				_offsetGain[1][2][j] = gain2[p];
				_offsetGain[1][3][j] = gain3[p];

				_offsetGain[2][0][j] = poly0[p];
				_offsetGain[2][1][j] = poly1[p];
				_offsetGain[2][2][j] = poly2[p];
				_offsetGain[2][3][j] = poly3[p];
			}
			break;
		}

	if (n>aduConv->rows())
		errstr<<"Could not find entry for CCD #"<<ccd<<"in `"
			  <<aduConv->qualifiedName()<<"'"
			  <<Sas::error("CcfDataIntegrityError");

	delete cdCcdId;
	delete cdOffset0;
	delete cdOffset1;
	delete cdOffset2;
	delete cdOffset3;
	delete cdGain0;
	delete cdGain1;
	delete cdGain2;
	delete cdGain3;
	delete cdPoly0;
	delete cdPoly1;
	delete cdPoly2;
	delete cdPoly3;

	//
	//	read FAST_G attribute - see email D. Lumb 2000-11-08
	//
	_fastG = 1.;
	if (aduConv->hasAttribute("FAST_G"))
		_fastG = aduConv->attribute("FAST_G")->asReal();

	//
	//	read energy combination coefficients
	//
	if (!_coefRead) {
		_coefRead = true;

		ColumnData *cdPattern = eCombine->column("PATTERN_ID")->data(),
				   *cdCoef = eCombine->column("ENERGY_PARAM")->data();
		int8 *pattern = cdPattern->int8Data();
		real32 *coef = cdCoef->real32Data();

		if (cdCoef->elements() != 4)
			errstr<<"Expecting 4 elements in column `"
				  <<cdPattern->parent()->qualifiedName()<<"'"
				  <<Sas::error("CcfDataIntegrityError");

		unsigned N = eCombine->rows();
		_e1Coef.resize(N);
		_e2Coef.resize(N);
		_e3Coef.resize(N);
		_e4Coef.resize(N);
		for (unsigned n=0; n<N; ++n) {
			unsigned patt = pattern[n];
			if (patt >= _e1Coef.size())
				errstr<<"Unexpected pattern ID "<<patt<<" in column `"
					  <<cdPattern->parent()->qualifiedName()<<"'"
					  <<Sas::error("UnknownPattern");
			_e1Coef[patt] = coef[cdCoef->position(n, 0)];
			_e2Coef[patt] = coef[cdCoef->position(n, 1)];
			_e3Coef[patt] = coef[cdCoef->position(n, 2)];
			_e4Coef[patt] = coef[cdCoef->position(n, 3)];
		}

		delete cdPattern;
		delete cdCoef;
	}
} // EmosEnergyCorrector::EmosEnergyCorrector()


//
//	see email D. Lumb, dated 2000-04-13
//
CalReal32Vector &EmosEnergyCorrector::gainCorrect(const CalReal64Vector &,
												  const CalReal32Vector &eIn,
												  const CalInt8Vector &pattern,
												  CalReal32Vector &eOut) const
{
	//
	//	fudge: for CCD1 in modes FastUncompressed/FastCompressed additional
	//	multiplicative factor has to be taken into account
	//
	float effFastG = 1.;
	Mode mode = calServer.state()->mode()->get();
	if (calServer.state()->ccd()->get() == 1 &&
		(mode == Xmm::FAST_UNCOMPRESSED || mode == Xmm::FAST_COMPRESSED))
		effFastG = _fastG;

	int nodeIndx = parent()->node();
	unsigned N = eIn.size();
	for (unsigned n=0; n<N; ++n) {
		unsigned patternIndx = (pattern[n]+3)/4;
		if (patternIndx > 3)
			patternIndx = 3;
		//
		//	pattern ID		patternIndx
		//	0				0
		//	1,2,3,4			1
		//	5,6,7,8			2
		//	9,10,11,12		3
		//	13,...			3	- treat all pattern >12 as pattern [9, 12]
		//

		eOut[n] = _offsetGain[0][patternIndx][nodeIndx] +
				  _offsetGain[1][patternIndx][nodeIndx] * eIn[n] +
				  _offsetGain[2][patternIndx][nodeIndx] * eIn[n] * eIn[n];
		eOut[n] *= effFastG;
	}

	return eOut;
} // EmosEnergyCorrector::gainCorrect()


//
//	--------------------------( EPN )--------------------------------------
//
bool EpnEnergyCorrector::_gainParamsSet = false;
EpnEnergyCorrector::EpnGainParameters EpnEnergyCorrector::_gainParams;
vector<float> EpnEnergyCorrector::_gainParamsSuppl;
vector<double> EpnEnergyCorrector::_Acoef,
			   EpnEnergyCorrector::_Bcoef;
double EpnEnergyCorrector::_adu_eV,
	   EpnEnergyCorrector::_rawy0;

EpnEnergyCorrector::EpnEnergyCorrector(Ccd ccd,
									   const CcfConstituentTable &aduConvTab,
									   const Table *gainHigh,
									   const Table *reEmission,
									   const Ccf *ccf,
							   CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent),
	  EnergyCorrector(ccd, aduConvTab, parent)
{
	if (!_gainParamsSet) {
		_gainParamsSet = true;

		if (gainHigh->rows() != N_PARAMETERS)
			errstr<<"Expected to find "<<N_PARAMETERS<<" gain parameters "
					"in PN high-accuracy gain table `"
				  <<gainHigh->qualifiedName()<<"' but there are only "
				  <<gainHigh->rows()
				  <<Sas::error("CcfDataIntegrityError");

		const Column *gainCol = gainHigh->column("GAIN_PAR");
		ColumnData *gainParCd = gainCol->data();
		unsigned nElements = gainParCd->elements();
		if (nElements != xmmPayLoad.epn.nNodes() * xmmPayLoad.epn.nCcds())
			errstr<<"Expected to find "<<xmmPayLoad.epn.nNodes() *
										 xmmPayLoad.epn.nCcds()
				  <<" elements in column "<<gainCol->qualifiedName()
				  <<" but there are "<<nElements
				  <<Sas::error("CcfDataIntegrityError");

		real32 *gainPar = gainParCd->real32Data();

		//
		//	allocate memory
		//
		for (unsigned n=0; n<N_PARAMETERS; ++n)
			_gainParams.push_back(new float [nElements]);

		for (unsigned n=0; n<N_PARAMETERS; ++n)
			for (unsigned pos=0; pos<nElements; ++pos)
				_gainParams[n][pos] = gainPar[gainParCd->position(n, pos)];

		delete gainParCd;

		//
		//	read supplemental gain parameters
		//
		_gainParamsSuppl.push_back(gainHigh->attribute("FINEPAR1")->asReal());
		_gainParamsSuppl.push_back(gainHigh->attribute("FINEPAR2")->asReal());
		_gainParamsSuppl.push_back(gainHigh->attribute("FINEPAR3")->asReal());

		//
		//	read re-emission threshold parameters
		//
		_adu_eV = reEmission->attribute("ADU_P_EV")->asReal();
		_rawy0 = reEmission->attribute("RAWY0")->asReal();
		CalOalUtility::fillVector(reEmission, "A", 0, _Acoef);
		CalOalUtility::fillVector(reEmission, "B", 0, _Bcoef);
		if (_Acoef.size() != 2 || _Bcoef.size() != 2)
			errstr<<"Expected vector columns A/B to be of size 2 in `"
				  <<reEmission->qualifiedName()<<"'"
				  <<Sas::error("CcfDataIntegrityError");
	}

    /* Get the current observing mode */
	unsigned short modeIndex = calServer.instrument()->
								modeIndex(calServer.state()->mode()->get());
	//
	//	read paramters for additional CTI/gain corrections from CTI CCF
	//
	const XmmInstrument *ins = &XmmPayLoad::epn;
	CifEntry *e = ccf->entry(ins, Ccf::CTI);
	DataSet *set = e->open();

	//	Orsay gain scaling
	Table *tab = set->table("ORSAY_GAIN");
	CalOalUtility::fillVector(tab, "ORSAY_GAIN", _orsayGain);
	if (_orsayGain.size() != ins->nCcds())
		errstr<<"Expected "<<ins->nCcds()<<" Orsay gain parameters in `"
			  <<tab->qualifiedName()<<"'"
			  <<Sas::error("CcfIntegrityError");

	//	long-term CTI increase
     /*  algoid=2 uses a ccd-dependent long-term correction - so have
         to make all long-term vectors per ccd */
	CcfConstituentTable ltCtiTab(set, "LONG_TERM_CTI", 5);
	const Table *lttab = ltCtiTab.block();

     /* Set the algoid as a variable for future use */
        _longAlgoid=ltCtiTab.algoid();

	if (ltCtiTab.algoid() == 0) {
		//	ALGOID=0 LT table structure, no-mode dependence; single vector
		//	column if size 5 (0: T-coef, 1-3: E-coef, 4: shift)
		vector<float> temp;
		CalOalUtility::fillVector(lttab, "LONG_TERM_CTI", temp);
		if (temp.size() != 5)
			errstr<<"Expected 5 long-term CTI parameters in `"
				  <<lttab->qualifiedName()<<"'"
				  <<Sas::error("CcfIntegrityError");

        for (unsigned c=1; c<=ins->nCcds(); ++c) {
		  _longTermCtiTcoef[c].assign(4, 0.);	//	3. order polynomial
		  _longTermCtiTcoef[c][1] = temp[0];	//	linear coeff is non-zero
		  _longTermCtiEcoef[c].assign(temp.begin()+1, temp.begin()+4);
		  _longTermCtiShift[c].push_back(temp[4]);
          _longTermCtiCcd.push_back(short(c));
        }

	} else if (ltCtiTab.algoid() == 1){

	   //	ALGOID=1 table structure: mode-dependence; three columns
	   //	with T/E coefficients + shift value
	    vector<short> modes;
  	    CalOalUtility::fillVector(lttab, "MODE_ID", modes);
	    vector<short>::const_iterator i = find(modes.begin(), 
                                                   modes.end(), modeIndex);
 	    if (i == modes.end())
		errstr<<"Did not find entry matching mode "
		  <<calServer.state()->mode()->get()
		  <<" in "<<lttab->qualifiedName()
		  <<Sas::error("CcfIntegrityError");

	    unsigned n=i-modes.begin();

           // Fill a structure for this mode for each CCD
           std::vector<float> tempT, tempE, tempS;
           CalOalUtility::fillVector(lttab, "T_COEFF", n, tempT);
           CalOalUtility::fillVector(lttab, "E_COEFF", n, tempE);
           CalOalUtility::fillVector(lttab, "SHIFT", n, tempS);

           for (unsigned c=1; c<=ins->nCcds(); ++c) {
             _longTermCtiTcoef.push_back(tempT);
             _longTermCtiEcoef.push_back(tempE);
             _longTermCtiShift.push_back(tempS);
             _longTermCtiCcd.push_back(short(c));
           }   

        } else if (ltCtiTab.algoid() >= 2) {

         /* ALGOID=2/3 table structure: mode-dependence, ccd-dependence;
                   three columns with T/E coefficients + shift value */
         /* ALGOID=4 table structure: mode, ccd, energy, Tcoeff(5), shift */
         /* ALGOID=5 table structure: mode, ccd, energy, Tcoeff(n), shift + 
                     LTC_TIMES - with n times */

          vector<short> ccdid, modes;
          CalOalUtility::fillVector(lttab, "CCD_ID", ccdid);
          CalOalUtility::fillVector(lttab, "MODE_ID", modes);

        /* Get the log energy to use for non-energy-interpolated - algo=4 ltcti */
          if (ltCtiTab.algoid() ==4)
             _longEBASE = log10( lttab->attribute("EBASE")->asReal() );

        // Loop over each entry and fill the vectors for the required mode
          short last_ccd=0;
          std::vector<float> tempT, tempE, tempS, tempEnergy;

          for (unsigned n=0; n<lttab->rows(); ++n)
          {
           if (modes[n] == modeIndex) {

             CalOalUtility::fillVector(lttab, "T_COEFF", n, tempT);
             CalOalUtility::fillVector(lttab, "SHIFT", n, tempS);
             _longTermCtiTcoef.push_back(tempT);
             _longTermCtiShift.push_back(tempS);
             _longTermCtiCcd.push_back(ccdid[n]);

            // algoid dependent terms
             if (ltCtiTab.algoid() <=3) {
                CalOalUtility::fillVector(lttab, "E_COEFF", n, tempE);
                _longTermCtiEcoef.push_back(tempE);
             } else {
                CalOalUtility::fillVector(lttab, "ENERGY", n, tempEnergy);
                _longTermCtiEnergy.push_back(tempEnergy);
             }

             // Check that CCDs are in order
             if (n>0) assert(ccdid[n]>=last_ccd);
             last_ccd=ccdid[n];
           }
          }

          // read in times for long_term correction, if needed
          if (ltCtiTab.algoid() ==5) {
              CcfConstituentTable ltTimTab(set, "LTC_TIMES");
              const Table *ltimtab = ltTimTab.block();

              CalOalUtility::fillVector(ltimtab, "TIME", 0, _longTermCtiTimes);
          }

          //cout<<"Siz: "<<_longTermCtiTcoef[0].size()<<" "<<_longTermCtiTimes.size()<<" "<<ltCtiTab.algoid()<<endl;
          // Check that the right number of long term coeffs were in the file
          if ( (_longTermCtiTcoef[0].size() != 4 && ltCtiTab.algoid() <=3)
             ||(_longTermCtiTcoef[0].size() != 5 && ltCtiTab.algoid() ==4)
             ||(_longTermCtiTcoef[0].size() != _longTermCtiTimes.size() &&
                ltCtiTab.algoid() ==5))
               errstr<<"Vectors in "<<lttab->qualifiedName()<<" do not have "
               "expected sizes"<<Sas::error("CcfIntegrityError");

        } // End of algoid>2

	//	CCD offsets
	tab = set->table("CCD_OFFSETS");
	CalOalUtility::fillVector(tab, "CCD_OFFSETS", _ccdOffsets);
	if (_ccdOffsets.size() != ins->nCcds())
	  errstr<<"Expected "<<ins->nCcds()<<" CCD offset parameters in `"
		<<tab->qualifiedName()<<"'"
		<<Sas::error("CcfIntegrityError");

	/*	Temperature gain variations.
	  If available take from QBOXTEMP structure */
	/* Also needed for the old (neves used) correction (AI)*/
	// Extract the gain slope for the required mode
	std::vector<float> gslope;

	if (set->hasBlock("QBOXTEMP_GAIN"))
	  {
	    tab = set->table("QBOXTEMP_GAIN");

	    vector<short> qt_ccdid, qt_modes;
	    CalOalUtility::fillVector(tab, "CCD_ID", qt_ccdid);
	    CalOalUtility::fillVector(tab, "MODE_ID", qt_modes);

	    CalOalUtility::fillVector(tab, "SLOPE", gslope);
	    /* Nomimal Quadrant Box  Temperature (AI)*/
	    // Extract the nominal values from the columns
	    std::vector<float> tnom;

	    CalOalUtility::fillVector(tab, "C_REF", tnom);
	    /* Nomimal Quadrant Box  Temperature (AI)*/

	    short last_ccd=0;
	    for (unsigned n=0; n<tab->rows(); ++n) {
	      if (qt_modes[n] == modeIndex) {
		//_TgainSlope.push_back(gslope[n]);
		_TgainSlope[qt_ccdid[n]]=gslope[n];
		/*Nominal Quadrant Box Temperature */
		//_CRef.push_back(tnom[n]);
		_CRef[qt_ccdid[n]]=tnom[n];

		// Check that CCDs are in order
		if (n>0) assert(qt_ccdid[n]>last_ccd);
		last_ccd=qt_ccdid[n];
	      }
	    }

	  } else { // Use TEMP_CTI block
	  tab = set->table("TEMP_CTI");
	  CalOalUtility::fillVector(tab, "OFFSET", _TgainOffset);
	  /* Uses map instead of vector (AI) */
	  //CalOalUtility::fillVector(tab, "SLOPE", _TgainSlope);
	  CalOalUtility::fillVector(tab, "SLOPE", gslope );

	  for (unsigned n=0; n<tab->rows(); ++n) {
	    _TgainSlope[n+1]=gslope[n];
	  }


	}
	// Temperature data common to both structures
	_aduRef = tab->attribute("ADU_REF")->asReal();

	// Make sure that all CCDs have temperature slope relation
	// Can't do this because SW, TI,BU modes only have one CCD !
	//if ( _TgainSlope.size() != ins->nCcds())
	//	 errstr<<"Expected "<<ins->nCcds()<<" values in columns of `"
	//		  <<tab->qualifiedName()<<"'"
	//		  <<Sas::error("CcfIntegrityError");

	//	timing mode fine gain
	vector<float> tg, bg, eg;
	tab = set->table("TIMING_GAIN");
	CalOalUtility::fillVector(tab, "TIMING_GAIN", tg);
	tab = set->table("BURST_GAIN");
	CalOalUtility::fillVector(tab, "BURST_GAIN", bg);
	if (tg.size() != 4 || bg.size() != 4)
		errstr<<"Expected "<<ins->nCcds()<<" values in `"
			  <<tab->qualifiedName()<<"' and preceding table"
			  <<Sas::error("CcfIntegrityError");
	_timingGain[0].a1 = tg[0];
	_timingGain[0].a2 = tg[1];
	_timingGain[0].x0 = tg[2];
	_timingGain[0].dx = tg[3];
	_timingGain[1].a1 = bg[0];
	_timingGain[1].a2 = bg[1];
	_timingGain[1].x0 = bg[2];
	_timingGain[1].dx = bg[3];

  /* Gain correction for EFF and FF mode are late additions to the CCF.
     For backwards compatibility only use if the arrays exist in the CCF */
    if (set->hasBlock("EFF_GAIN"))
    {
	    CcfConstituentTable efgTab(set, "EFF_GAIN", 1);
		tab = set->table("EFF_GAIN");

     /* Read in the EFF_GAIN parameters depending on the ALGOID */
        if (efgTab.algoid() == 0) { // Just one set of params
	  	   CalOalUtility::fillVector(tab, "EFF_GAIN", eg);
        } else { // algoid = 1; Params for each CCD
           vector<unsigned short> ccdid;
           CalOalUtility::fillVector(tab, "CCD_ID", ccdid);

         /* Find the record for this CCD and read the gain values */
           vector<unsigned short>::const_iterator i = find(ccdid.begin(), ccdid.end(),
                                               ccd);
           if (i == ccdid.end())
            errstr<<"Did not find entry matching mode ccd:"
                  <<ccd<<" in "<<tab->qualifiedName()
                  <<Sas::error("CcfIntegrityError");

           unsigned n=i-ccdid.begin();
           CalOalUtility::fillVector(tab, "EFF_GAIN", n, eg);
        }
		_timingGain[2].a1 = eg[0];
		_timingGain[2].a2 = eg[1];
		_timingGain[2].x0 = eg[2];
  		_timingGain[2].dx = eg[3];
    	_effGainRead=true;
        //cout<< "Gain params: "<<eg[0]<<" "<<eg[1]<<" "<<eg[2]<<" "<<eg[3]
          //  <<endl;
	} else {
 		_effGainRead=false;
  	}

    if (set->hasBlock("FF_GAIN"))
    {
		tab = set->table("FF_GAIN");
		CalOalUtility::fillVector(tab, "FF_GAIN", eg);
		_timingGain[3].a1 = eg[0];
		_timingGain[3].a2 = eg[1];
		_timingGain[3].x0 = eg[2];
  		_timingGain[3].dx = eg[3];
    	_ffGainRead=true;
	} else {
 		_ffGainRead=false;
  	}


   /* Rate-dependent CTI correction:
    *
    *   If extension and data exists in the CCF then set the
    *   rate-dep CTI parameters, a0, a1 and a2 from the file.
    *   Otherwise set to 1.0,0.0,0.0 - which is the NULL correction
   */
    //vector<float> _rdc_a[3];
    if (set->hasBlock("RATE_DEPENDENT_CTI"))
    {
       tab = set->table("RATE_DEPENDENT_CTI");

       vector<short> rdc_modes;
       vector<float> vrdc_a0, vrdc_a1, vrdc_a2;

       CalOalUtility::fillVector(tab, "MODE_ID", rdc_modes);
       CalOalUtility::fillVector(tab, "A0", vrdc_a0);
       CalOalUtility::fillVector(tab, "A1", vrdc_a1);
       CalOalUtility::fillVector(tab, "A2", vrdc_a2);

      // Search for the right mode and set coeffs
       for (unsigned n=0; n<tab->rows(); ++n) {
           if (rdc_modes[n] == modeIndex) {
              if (_rdc_a.size() == 0) {  // Set the first time
                 _rdc_a.push_back(vrdc_a0[n]);
                 _rdc_a.push_back(vrdc_a1[n]);
                 _rdc_a.push_back(vrdc_a2[n]);
              }
            }
       }
    }

   // Push back dummy coeffs which result in no change in case none
   // were set
    if (_rdc_a.size() == 0) {
       _rdc_a.push_back(1.0);
       _rdc_a.push_back(0.0);
       _rdc_a.push_back(0.0);
    }

   /* Spatial CTI correction:
    *
    *   Based on polynomial fits to each column of each CCD
   */
    if (set->hasBlock("SPATIAL_CTI"))
    {
       tab = set->table("SPATIAL_CTI");

       vector<short> ccdid, modes;
       CalOalUtility::fillVector(tab, "CCD_ID", ccdid);
       CalOalUtility::fillVector(tab, "MODE_ID", modes);

     // Read in the parameters and templates for the required mode
       std::vector<float> tempP, tempT;

       short last_ccd=0;

       for (unsigned n=0; n<tab->rows(); ++n)
 	   {
	      if (modes[n] == modeIndex) {
	         CalOalUtility::fillVector(tab, "PARAM", n, tempP);
	         CalOalUtility::fillVector(tab, "TEMPLATE", n, tempT);
             _spatCtiPar.push_back(tempP);
             _spatCtiTmpl.push_back(tempT);
             _spatCtiCcd.push_back(ccdid[n]);

	     // Check that CCDs are in order - is this necessary ?
	        if (n>0) assert(ccdid[n]>last_ccd);
	           last_ccd=ccdid[n];
   	      }
    	  _spCtiRead=true;
       }
    }
      else { _spCtiRead=false; 
    }

   /* Rate-dependent PHA correction
    *
    *   Calibrated for Timing and Burst modes
    *
   */
    _rdp_burst_support=false;

     // Old block name
    if (set->hasBlock("RDPHA_DERIV"))
    {
       //tab = set->table("RDPHA_DERIV");
       readRDPHAcoeffs(set->table("RDPHA_DERIV"));
     // New names
    } else if (set->hasBlock("RDPHA_TIMING")) {
       readRDPHAcoeffs(set->table("RDPHA_TIMING"));
       readRDPHAcoeffs(set->table("RDPHA_BURST"));
       _rdp_burst_support=true;
    }

   // Read in the combined event offsets
    _combEvRead=false;

    if (set->hasBlock("COMB_EVT_OFFSET"))
    {
       tab = set->table("COMB_EVT_OFFSET");

       vector<short> modes, ccdid, patternId;
       vector<float> tempEnergy, tempshift;
       CalOalUtility::fillVector(tab, "MODE_ID", modes);
       CalOalUtility::fillVector(tab, "CCD_ID", ccdid);
       CalOalUtility::fillVector(tab, "PATTERN", patternId);
       CalOalUtility::fillVector(tab, "ENERGY", tempEnergy);
       CalOalUtility::fillVector(tab, "SHIFT", tempshift);

     // Read in the parameters and templates for the required mode
       _combEvMaxPatt=0;
       for (unsigned n=0; n<tab->rows(); ++n)
          {
    	     if (modes[n] == modeIndex) {
    	       _combEvCcd.push_back(ccdid[n]);
    	       _combEvPatt.push_back(patternId[n]);
    	       _combEvEnergy.push_back(tempEnergy[n]);
    	       _combEvShift.push_back(tempshift[n]);

    	      // Keep a track of the maximum pattern
    	       if (patternId[n] > _combEvMaxPatt) _combEvMaxPatt=patternId[n];
    	     }
          }
       _combEvRead=true;
    }

   // Read in the discarded line relationship
    if (set->hasBlock("NDSCLIN_GAIN"))
    {
       tab = set->table("NDSCLIN_GAIN");

       vector<short> modes, ccdid;
       vector<float> offset, slope;
       CalOalUtility::fillVector(tab, "MODE_ID", modes);
       CalOalUtility::fillVector(tab, "CCD_ID", ccdid);
       CalOalUtility::fillVector(tab, "OFFSET", offset);
       CalOalUtility::fillVector(tab, "SLOPE", slope);

       // Read in the parameters and templates for the required mode
       for (unsigned n=0; n<tab->rows(); ++n)
            {
                if (modes[n] == modeIndex) {
                   _ndsclinCCD.push_back(ccdid[n]);
                   _ndsclinOffset.push_back(offset[n]);
                   _ndsclinSlope.push_back(slope[n]);
                }
            }

       // Read in the conversion factor for the number of discarded lines
       // when MIPSEL=2
       _ndsclinMip2Conv = tab->attribute("MIPSEL2C")->asReal();

       // Read in the conversion factor for the number of discarded lines
       // when MAXMIP=5
       _ndsclinMaxMip5Conv = tab->attribute("MAXMIP5C")->asReal();

       _ndsclinGainRead=true;

    } else {
    	_ndsclinGainRead=false;
    }



} // EpnEnergyCorrector::EpnEnergyCorrector()


CalReal32Vector &EpnEnergyCorrector::gainCorrect(const CalReal64Vector &time,
											   const CalInt16Vector &rawX,
											   const CalInt16Vector &rawY,
											   const CalReal32Vector &eIn,
											   CalReal32Vector &eOut) const
{
	if (calServer.state()->accuracyLevel()->get() == LOW) {
		//
		//	low-accurcy gain correction
		//
		EnergyCorrector ec = *this;
		unsigned N = rawX.size();
		CalInt8Vector nodeVec(0, N);
		for (unsigned n=0; n<N; ++n)
			nodeVec[n] = rawX[n]-1;
		ec.gainCorrect(time, nodeVec, eIn, eOut);
		return eOut;
	}

	static bool first = true;

	if (first) {
		first = false;
		//
		//	first time called - need to initialize the PN gain corrector
		//	algorithm
		//
		unsigned N = xmmPayLoad.epn.nNodes() * xmmPayLoad.epn.nCcds();
		float *pp0 = _gainParams[0],
			  *pp1 = _gainParams[1],
			  *pp2 = _gainParams[2],
			  *pp3 = _gainParams[3],
			  *aa = _gainParams[6],
			  *bb = _gainParams[8];
		int M = _gainParamsSuppl.size();
		float suppl[M];
		for (int m=0; m<M; ++m)
			suppl[m] = _gainParamsSuppl[m];

		CALPNALGO_initGain(N, pp0, pp1, pp2, pp3, aa, bb, M, suppl);
	}

	bool mediumAccuracy = calServer.state()->accuracyLevel()->get() == MEDIUM;
	unsigned N = eIn.size();
	Ccd	ccd = parent()->ccd();
	unsigned quadId = EPNcam::toQuadrantId(ccd),
			 ccdId = EPNcam::toCcdId(ccd);
	for (unsigned n=0; n<N; ++n) {
		float	ampr = eIn[n];
		int 	x = INT(rawX[n]) - 1,
				y = INT(rawY[n]) - 1,
				ierr = 0;
		float gainCorrectedAmp, gainCtiCorrectedAmp;

		CALPNALGO_gainCorrect(quadId, ccdId, x, y, ampr, gainCorrectedAmp,
							  gainCtiCorrectedAmp, ierr);

		if (!ierr)
			eOut[n] = mediumAccuracy ? gainCtiCorrectedAmp : gainCorrectedAmp;
		else {
			errstr<<"Gain correction algorithm for event #"<<n
				  <<" (CCD/rawX = "<<ccd<<"["<<quadId<<"/"<<ccdId<<"]/"
				  <<rawX[n]<<") failed - error code: "<<ierr
				  <<" no correction applied - reason for failure: ";
			switch (ierr) {
				case 3: errstr<<"negative amplitude"; break;
				default: errstr<<"unknown"; break;
			}
			errstr<<Sas::warning("gainCorrectionFailed");
			eOut[n] = eIn[n];
		}
	}

	return eOut;
} // EpnEnergyCorrector::gainCorrect()


std::vector<double> EpnEnergyCorrector::additionalGainCtiCorrect(
											CalReal32Vector &ampl,
											const CalInt16Vector &rawY,
											double T,
											bool doTCorr,
											bool doOffsetCorr,
											bool doLongTermCTISoftCorr,
											bool doLongTermCTIyCorr,
											bool doLongTermCTICorr) const
{
	unsigned N = ampl.size();
	vector<double>	corrPar(12);
	const CalState *cs = calServer.state();
	double t = (cs->exposureStartTime()->get() - STime(2000U) +
				.5 * cs->exposureDuration()->get()) /
				STime::secondsPerNonLeapYear;	//	years since 01-01-2000 at
												//	middle of exposure

	Ccd ccd = cs->ccd()->get();
	corrPar[0] = _orsayGain[ccd-1];
	corrPar[1] = t;
	corrPar[2] = _aduRef;
	corrPar[7] = T;
        corrPar[3] = 0.0;
        for (unsigned i=8; i<=11; ++i) {corrPar[i]=0.0;}

	//
	//	energy rescaling compared to Orsay calibration
	//
	for (unsigned n=0; n<N; ++n)
		ampl[n] *= _orsayGain[ccd-1];

    //    First find the entry for this CCD
    short sccd = ccd;

    vector<short>::const_iterator c =
        find(_longTermCtiCcd.begin(), _longTermCtiCcd.end(), sccd);

    if (c == _longTermCtiCcd.end())
       errstr<<"Did not find entry matching mode ccd:"
             <<ccd<<" in Long-term CTI correction"
             <<Sas::error("CcfIntegrityError");

    unsigned cindex = c - _longTermCtiCcd.begin();

	//
	//	correct for long-term CTE decrease (modeled as gain)
	//
	double			lty0 = _longTermCtiShift[cindex][0];

    //cout<<"longterm: "<<ccd<<" "<<cindex<<" "<<_longTermCtiTcoef[cindex][2]<<endl;

    // Should a correction for long-term CTI be applied ?        
    if (doLongTermCTICorr) {

      // Calculate tau, the CTI effect per pixel at the time of the
      // observation and at the beginning of the mission (tau0)
       double tau=0.0, tau0=0.0, tt=0.0;
       //InterpolatedCurve tauEcurve=0, tau0Ecurve=0;
       Slatec::LinearInterpolant<double> *tauEcurve, *tau0Ecurve;
       if (_longAlgoid<=3) 
       {
          tau = _longTermCtiTcoef[cindex][0] +
                _longTermCtiTcoef[cindex][1] * t +
                _longTermCtiTcoef[cindex][2] * t*t +
                _longTermCtiTcoef[cindex][3] * t*t*t;

          tau0 = _longTermCtiTcoef[cindex][0];
       } 
       else // Algoid = 4 or 5
       {
         // Make an energy-dependent curve to interpolate with the tau 
         // values at the calibration energy points           
          vector<double> tauv, tauv0, eev;

         // Set a point at 10 eV with value tau=0
          eev.push_back(-2.0);
          tauv.push_back(0.0); tauv0.push_back(0.0);

         // Find the index of the long_term_CTI array for this times (algoid=5)
          double tfrac=0.0;
          unsigned t1=0, t2=0; // Index of long_term_cti array for time t 
          if (_longAlgoid==5) {
            // Find the index for the time just smaller than t
             for(unsigned i=0; i<_longTermCtiTimes.size(); ++i)
                if (t>=_longTermCtiTimes[i]) t1=i;
            // If t within the ranges of the tabultaed times - interpolate
             if (t1<(_longTermCtiTimes.size()-1) && t>=_longTermCtiTimes[0]) {
                t2=t1+1;
                tfrac = (t-_longTermCtiTimes[t1]) / 
                         (_longTermCtiTimes[t2]-_longTermCtiTimes[t1]);
             } else { // Use last CTI value if t outside times in array 
                t2=t1;
             }
          }
             
         // Loop over the CCF entries for this mode 
          for(unsigned c=0; c<_longTermCtiCcd.size(); ++c){

            // Process entry if from the right CCD
             if (_longTermCtiCcd[c] == sccd)
              {

               // Algo 4 with coefficients
                if (_longAlgoid==4) {

                 // Calculate tau for this energy and time
                  tauv.push_back(_longTermCtiTcoef[c][0] +
                                 _longTermCtiTcoef[c][1] * t +
                                 _longTermCtiTcoef[c][2] * t*t +
                                 _longTermCtiTcoef[c][3] * t*t*t +
                                 _longTermCtiTcoef[c][4] * t*t*t*t );

                } else if (_longAlgoid==5) {

                 // Interpolate in time
                  tauv.push_back(_longTermCtiTcoef[c][t1] + 
                      ( _longTermCtiTcoef[c][t2] - 
                               _longTermCtiTcoef[c][t1] ) * tfrac );

           /* debug
                  double t3 = (_longTermCtiTcoef[c][t1] +
                      ( _longTermCtiTcoef[c][t2] -
                               _longTermCtiTcoef[c][t1] ) * tfrac);
                  cout<<"..."<<t3<<endl;
              end debug */

                }
              
                tauv0.push_back(_longTermCtiTcoef[c][0]);
                eev.push_back(log10(_longTermCtiEnergy[c][0]));
              }
          }
          //cout<<"Res: "<<_longTermCtiTcoef[0][10]<<" "<<t<<" "<<t1<<" "<<t2<<" "<<tfrac<<" "<<tauv[1]<<endl;
         /* Add a last point so that the extrapolation beyond the final
          * calibrated energy point is flat.
          */
          eev.push_back(eev[eev.size()-1]+0.1);
          tauv.push_back(tauv[tauv.size()-1]); tauv0.push_back(tauv0[tauv0.size()-1]);

          // Make the interpolation curves - SHOULD BE LINEAR
//          tauEcurve = InterpolatedCurve(eev, tauv);
//          tau0Ecurve = InterpolatedCurve(eev, tauv0);
          tauEcurve = new Slatec::LinearInterpolant<double>(eev, tauv);
          tau0Ecurve = new Slatec::LinearInterpolant<double>(eev, tauv0);

          //cout<<"TAUV:"<<tauv[0]<<" "<<tauv[1]<<tauv[2]<<endl;
          //cout<<"TAUV0:"<<tauv0[0]<<" "<<tauv0[1]<<endl;
          //cout<<"EEV:"<<eev[0]<<" "<<eev[1]<<endl;
       }
          // cout<<"C: "<<_longTermCtiTcoef[0][1];

      // Loop over events and do the correction
       for (unsigned i=0; i<N; ++i) 
        {
         // Find log of energy of this event
          double logkeV = log10(ampl[i]*5.e-3);

         // Calculate tt, the numbe of CTI shifts
          if (_longAlgoid<=3 && doLongTermCTISoftCorr) 
           {
             tt = _longTermCtiEcoef[cindex][0] +
                  logkeV*(_longTermCtiEcoef[cindex][1] +
                  _longTermCtiEcoef[cindex][2]*logkeV);

             if (doLongTermCTIyCorr) tt *= rawY[i]/lty0;

             corrPar[8] = _longTermCtiEcoef[cindex][0];
             corrPar[9] = _longTermCtiEcoef[cindex][1];
             corrPar[10] = _longTermCtiEcoef[cindex][2];

           }
          else // All other options
           {
             tt = doLongTermCTIyCorr ? rawY[i]/lty0 : 1.;
           }
  
         // Make the correction to the amplitude
          if (_longAlgoid<=2 || (_longAlgoid==3 && doLongTermCTISoftCorr))
           {
             ampl[i] *= 1. + tau*tt;
           }
          else // All other options
           {

            /* If algo=4 or 5 then interpolate in log energy to find 
               tau and tau0 for this event. If energy interpolation is
	       not required then use the Mn values (SOMEHOW!) */
             if (_longAlgoid>=4) {
                if (doLongTermCTISoftCorr)
                 {
                   tau0 = (*tau0Ecurve)(logkeV);
                   tau = (*tauEcurve)(logkeV);

                 } else {
                   tau0 = (*tau0Ecurve)(_longEBASE);
                   tau = (*tauEcurve)(_longEBASE);
                 }
             }
             //cout<<"FTT "<<logkeV<<" "<<tau<<" "<<tau0<<endl;

             ampl[i] *= pow( ((1.0 - tau0) / (1. - tau )), tt ); 
           }
        } // end of loop over events

       /* Set corrPar values for later storage */
        corrPar[3] = _longTermCtiTcoef[cindex][1];
        corrPar[11] = doLongTermCTIyCorr ? lty0 : 0.;

    } // End of If long-term CTI block


	//
	//	correct for CCD offset (modeled as shift):
	//
	if (doOffsetCorr) {
		for (unsigned i=0; i<N; ++i)
			ampl[i] -= _ccdOffsets[ccd-1];
		corrPar[4] = _ccdOffsets[ccd-1];
	} else
		corrPar[4] = 0.;

	//
	//	correct for temperature variations (modeled as gain):
	//
	if (doTCorr) {
	  //map<int,float>::const_iterator it;   - MJF - not used
	  map<int,float>::const_iterator CRefIT;
	  map<int,float>::const_iterator slopeIT;
	  for (unsigned i=0; i<N; ++i)
	    {
	      /* Change to new linear only (no offset) calculation.
		 The original calculation was never used, so no need to
		 make backwards compatible. */

	      //ampl[i] *= (1. - T*_TgainSlope[ccd-1]) / _aduRef;

	      /*Now uses maps (AI) */
	      //ampl[i] *= (1. - (T-_CRef[ccd-1])*_TgainSlope[ccd-1]);
	      CRefIT  = _CRef.find(ccd);
	      slopeIT = _TgainSlope.find(ccd);
	      //cout<<"AMPL BEFORE: "<<ampl[i]<<" "<<T<<" "<< CRefIT->second << " "<< slopeIT->second <<endl;
	      /* Final modification. Instead of a temperature they want a constant value (AI) */
	      //ampl[i] *= (1. - (T-tempIT->second)*slopeIT->second);
	      ampl[i] *= (1. - T*slopeIT->second)*CRefIT->second;
	      //cout<<"AMPL AFTER: "<<ampl[i]<< endl;
	    }
	  //ampl[i] *= (1. - (_TgainOffset[ccd-1] + T*_TgainSlope[ccd-1]))/
	  //	_aduRef;
	  corrPar[5] = 0.;
	  //corrPar[5] = _TgainOffset[ccd-1];
	  /* now maps (AI)*/
	  //corrPar[6] = _TgainSlope[ccd-1];
	  corrPar[6] = slopeIT->second;

	  errstr<<"   Temperature for gain [deg C]  = "<<T<<endl
		<<"   Temperature slope [adu/K]     = "<< slopeIT->second <<endl
		<<"   Temperature reference channel = "<<_aduRef<<endl;


	  /*Just in case the old algorithm is used or not. We
	    have to check that the vector has been filled */
	  if (_CRef.size() != 0)
	    errstr<<"  correction factor to adjust peak position  = " <<  CRefIT->second  <<endl;

	  if (_TgainOffset.size() != 0)
	    errstr <<"   Temperature offset [adu]  = "<<_TgainOffset[ccd-1] <<endl;

	  errstr << Sas::message(Msg::App, Msg::Sparse);
	  /* Print the diagnostic info (AI) */
	} else {
	  corrPar[5] = 0.;
	  corrPar[6] = 0.;
	}

	//	diagnostic output
	if (errHandler.verbosity(Msg::App) >= Msg::Verbose) {
	  errstr<<' '<<Sas::message(Msg::App, Msg::Sparse);
	  errstr<<"   number of CCD to be corrected = "<<ccd<<endl
		<<"   Orsay gain scaling            = "<<_orsayGain[ccd-1]<<endl
		<<"   Years elapsed since 1.1.2000  = "<<t<<endl
		<<Sas::message(Msg::App, Msg::Sparse);
	  if (doLongTermCTICorr) {
	    errstr<<"   CTI increase [1.E-5/yr]       = "
		  <<_longTermCtiTcoef[cindex][1]*1.e5<<endl;
	    if (doLongTermCTISoftCorr && _longAlgoid<=3) { // Energy coeffs
	      errstr<<"   energy dependence: coeff(0)  = "
		    <<_longTermCtiEcoef[cindex][0]<<endl
		    <<"   energy dependence: coeff(1)  = "
		    <<_longTermCtiEcoef[cindex][1]<<endl
		    <<"   energy dependence: coeff(2)  = "
		    <<_longTermCtiEcoef[cindex][2]<<endl
		    <<Sas::message(Msg::App, Msg::Sparse);
	    }
	    errstr<<"   reference Y shifts            = "<<lty0<<endl
		  <<Sas::message(Msg::App, Msg::Sparse);
	  }
	  if (doOffsetCorr)
	    errstr<<"   CCD offset [adu]              = "<<_ccdOffsets[ccd-1]
		  <<endl<<Sas::message(Msg::App, Msg::Sparse);

	}

	return corrPar;
} // EpnEnergyCorrector::additionalGainCtiCorrect()


void EpnEnergyCorrector::bkgDepGainCorrect(
	CalReal32Vector &ampl, 
	int ndsclin, 
        short mipsel,
	short maxmip) const
{
 // Perform correction for the background dependence of the gain if the
 // params were in CCF
  if (_ndsclinGainRead) {

   // For the moment can only process data with mipsel=1 or 2 and
   // maxmip=5 or 63
    if (mipsel == 1 || mipsel ==2) {

      if (maxmip == 63 || maxmip ==5) {

       // Use a conversion factor for ndsclines if mipsel=2
        double ndsclines=double(ndsclin);
        if (mipsel == 2)
           ndsclines = double(ndsclin) / _ndsclinMip2Conv;
 
       // Use a conversion factor for ndsclines if maxmip=5
        if (maxmip == 5)
           ndsclines =  ndsclines / _ndsclinMaxMip5Conv;

   //cout<<"ndsclines: "<<ndsclines<<endl;
  	unsigned N = ampl.size();

	// Find the index for this CCD
	const CalState *cs = calServer.state();
	Ccd ccd = cs->ccd()->get();
	short sccd = ccd; // The current CCD

	short cindex=-1;
        for (unsigned i = 0; i< _ndsclinCCD.size(); ++i){
      	   if ( _ndsclinCCD[i] == sccd)
              cindex=i;
        }

        if (cindex<0)
           errstr<<"Did not find entry matching mode ccd:"
                 <<ccd<<" in NDSCLINE gain correction"
                 <<Sas::error("CcfIntegrityError");

      // Loop over events and do the correction
        for (unsigned i=0; i<N; ++i) {

         // The amplitude is multiplied by a linear function of the number
         // of discarded lines, which is a proxy for the background level.
     	  ampl[i] *= (1 + _ndsclinOffset[cindex] + 
                              _ndsclinSlope[cindex] * ndsclines);
        }

      } else { // MAXMIP!=5 or 63 can't be handled
     	      errstr<<"The background-dependent gain correction cannot be "
	          "calculated when MAXMIP is not 5 or 63"
		  <<Sas::warning("BCKgainCorrectionMAXMIP");
      }
    } else { // MIPSEL!=1 or 2 can't be handled
	errstr<<"The background-dependent gain correction cannot be calculated "
			"when MIPSEL is not 1 or 2"
		  <<Sas::warning("BCKgainCorrectionMIPSEL12");
    }
  } else { // No gain correction parameters found in CCF
 	 errstr<<"Background (discarded line) correction parameters were not "
	 	 "present for PN in the CCF. No correction will be applied "
		 <<Sas::warning("NoBCKgainCorrectionFound");
  }
} // EpnEnergyCorrector::bkgDepGainCorrect()


void EpnEnergyCorrector::spatialCtiCorrect(

        CalReal32Vector &ampl,

        const CalInt16Vector &rawX,

        const CalInt16Vector &rawY) const
{

    // Perform correction if the spatial correction params were in CCF
      if ( _spCtiRead ) 
      {       
	int N = ampl.size();

       // Hardwire the number of expected parameters
        unsigned npar=3;
        //unsigned ncol=64;   - MJF - not used
        unsigned nrow=200;

       // Find the right parameters for this CCD
	const CalState *cs = calServer.state();
	Ccd ccd = cs->ccd()->get();
        
        short sccd = ccd;

        vector<short>::const_iterator cIt =
          find(_spatCtiCcd.begin(), _spatCtiCcd.end(), sccd);

        if (cIt == _spatCtiCcd.end())
             errstr<<"Did not find entry matching mode ccd:"
             <<ccd<<" in Spatial CTI correction"
             <<Sas::error("CcfIntegrityError");

        unsigned cindex = cIt - _spatCtiCcd.begin();

       // Loop over each event
        double a=0.0, b=0.0, c=0.0, h=0.0, f=0.0;
	for (int n=0; n<N; ++n)
         {

       // Apply the formula if amplitude > 0
           if ( ampl[n] > 0.0 ) 
           {

            int start = npar * (rawX[n]-1); 
            a = _spatCtiPar[cindex][start];
            b = _spatCtiPar[cindex][start+1];
            c = _spatCtiPar[cindex][start+2];

            h = log10(ampl[n]) - 3.0;
            f = a + (b + c * h) * h;
            f = max(f,0.0);

            //cout<<"Idata: "<<ccd<<" "<<rawX[n]<<" "<<rawY[n]<<" "<<_spatCtiTmpl[cindex][0]<<" "<<_spatCtiTmpl[cindex][1]<<endl;
            //cout<<"PARAMS: "<<a<<" "<<b<<" "<<c<<" "<<_spatCtiTmpl[cindex][nrow * (rawX[n]-1) + (rawY[n]-1)]<<endl;
            //cout<<"RESULT: "<<h<<" "<<f<<" "<<_spatCtiTmpl[cindex][nrow * (rawX[n]-1) + (rawY[n]-1)] * f<<endl;

	    ampl[n] -= _spatCtiTmpl[cindex][nrow * (rawX[n]-1) + 
                                                      (rawY[n]-1)] * f;
           }

         }
    // DEBUG
        //int ii=N-1;
        //cout<<"Correcting: "<<N<<" events. "<<sccd<<" "<<cindex<<" "<< a<<" "<<b<<" "<<c<<" "<<h<<" "<<f<<" "<<rawX[ii]<<" "<<rawY[ii]<<" "<<ampl[ii]<<" "<<_spatCtiTmpl[cindex][nrow * (rawX[ii]-1) + (rawY[ii]-1)]<<endl;
    // DEBUG

      } else {
             errstr<<"Did not find pn spatial CTI parameters in CCF -"
                   <<" not applying Spatial CTI correction"
                   <<Sas::warning("NoSpatialCTICorrectionFound");
      }

} // EpnEnergyCorrector::spatialCtiCorrect

void EpnEnergyCorrector::combinedEventCorrect(

        CalReal32Vector &ampl,
        const CalInt16Vector &patternId) const
{

  // Corrected for the combined event offset if COMB_EVT parameters in CCF
    if (_combEvRead)
    {
     // Create an energy interpolater for the required CCD and for each pattern
 	  const CalState *cs = calServer.state();
	  Ccd ccd = cs->ccd()->get();
	  short sccd = ccd;

     // Find max number of possible patterns
	  short Np = _combEvMaxPatt+1;

     // Initialise vectors for log_energy and shift for all patterns
     // Set an initial point at 10 eV
      vector<vector<double> > eev(Np), shift(Np);
      for (int p=0; p<Np; ++p){
            eev[p].push_back(-2.0); shift[p].push_back(0.0);
      }

     // Loop over each CCF entry and extract the ones for this CCD
      for (unsigned c=0; c<_combEvCcd.size(); ++c){

        // Process entry if from the right CCD
         if (_combEvCcd[c] == sccd)
          {
           shift[_combEvPatt[c]].push_back(_combEvShift[c]);
           eev[_combEvPatt[c]].push_back(log10(_combEvEnergy[c]));
          }
      }

     // Create an interpolator in energy for each pattern
      vector<Slatec::LinearInterpolant<double> *> shiftEcurve;

     // Add a final point to give a linear extrapolation at high E
     // and make the interpolators for patterns which have CCF data
      for (int p=0; p<Np; ++p) {
    		eev[p].push_back(eev[p][eev[p].size()-1]+0.1);
    		shift[p].push_back(shift[p][shift[p].size()-1]);
    		shiftEcurve.push_back(new
    			      Slatec::LinearInterpolant<double>(eev[p], shift[p]));
      }

     // Correct the energy of each event for the shift due to pattern
     // recombination
      int N = ampl.size();

      for (int n=0; n<N; ++n)
        {
         // Perform correction if energy is +ve. This explicitely excludes NULLs
          if (ampl[n]>0) {

    	   // Find the shift for this pattern and energy
           // Subtract the combination shift. Ignore too large patterns
    	    if (patternId[n]>=0 && patternId[n]<=_combEvMaxPatt) // slow !
            {
              // Calculate the log(keV) of this amplitude - input in eV
    	       double logkeV = log10(ampl[n]*1.e-3);

              // Calculate the new amplitude
               ampl[n] -= (*shiftEcurve[patternId[n]])(logkeV);
            }
          }
        }

    } // End of COMB EVT CCF block read ok

} // EpnEnergyCorrector::combinedEventCorrect

void EpnEnergyCorrector::gainTimingCorrect(CalReal32Vector &ampl) const
{

	unsigned N = ampl.size();

	//
	//	fine-adjustment of the gain in TIMING mode to be applied after all
	//	the gain corrections or immediately before the CTI corrections
	//	the input to the routine is in [adu], the coefficients have been
	//	derived for energies in [eV], therefore scale input by
	//	1 adu = 5.e-0 keV
	//
	double a0 = _timingGain[0].a1 - _timingGain[0].a2;
	for (unsigned n=0; n<N; ++n) {
		double eeV = ampl[n]*5.e-0,
			   h = _timingGain[0].a2 + a0/(1.+exp((eeV-_timingGain[0].x0)/
												_timingGain[0].dx));
		ampl[n] /= h;
	}
} // EpnEnergyCorrector::gainTimingCorrect()


void EpnEnergyCorrector::gainBurstCorrect(CalReal32Vector &ampl) const
{

	unsigned N = ampl.size();

	//
	//	fine-adjustment of the gain in Burst mode to be applied after all
	//	the gain corrections or immediately before the CTI corrections
	//	the input to the routine is in [adu], the coefficients have been
	//	derived for energies in [eV], therefore scale input by
	//	1 adu = 5.e-0 keV
	//
	double a0 = _timingGain[1].a1 - _timingGain[1].a2;
	for (unsigned n=0; n<N; ++n) {
		double eeV = ampl[n]*5.e-0,
			   h = _timingGain[1].a2 + a0/(1.+exp((eeV-_timingGain[1].x0)/
												_timingGain[1].dx));
		ampl[n] *= h;
	}
} // EpnEnergyCorrector::gainBurstCorrect()

void EpnEnergyCorrector::gainEffCorrect(CalReal32Vector &ampl) const
{

    // If Eff gain correction parameters have been read from
    // the CCF then correct the gain. Otherwise issue a warning.
    if (_effGainRead) {

  		unsigned N = ampl.size();

		//
		//	fine-adjustment of the gain in Eff mode to be applied after all
		//	the gain corrections or immediately before the CTI corrections
		//	the input to the routine is in [adu], the coefficients have been
		//	derived for energies in [eV], therefore scale input by
		//	1 adu = 5.e-0 keV
		//
		double a0 = _timingGain[2].a1 - _timingGain[2].a2;
		for (unsigned n=0; n<N; ++n) {
			double eeV = ampl[n]*5.e-0,
			   h = _timingGain[2].a2 + a0/(1.+exp((eeV-_timingGain[2].x0)/
												_timingGain[2].dx));
			ampl[n] *= h;
		}

    } else { // No gain correction parameters found in CCF
		errstr<<"Gain correction parameters were not present for PN "
                "EFF mode in the CCF. No correction will be applied "
			  <<Sas::warning("NoEFFgainCorrectionFound");
    }

} // EpnEnergyCorrector::gainEffCorrect()

void EpnEnergyCorrector::gainFfCorrect(CalReal32Vector &ampl) const
{

    // If Eff gain correction parameters have been read from
    // the CCF then correct the gain. Otherwise issue a warning.
    if (_ffGainRead) {

  		unsigned N = ampl.size();

		//
		//	fine-adjustment of the gain in Eff mode to be applied after all
		//	the gain corrections or immediately before the CTI corrections
		//	the input to the routine is in [adu], the coefficients have been
		//	derived for energies in [eV], therefore scale input by
		//	1 adu = 5.e-0 keV
		//
		double a0 = _timingGain[3].a1 - _timingGain[3].a2;
		for (unsigned n=0; n<N; ++n) {
			double eeV = ampl[n]*5.e-0,
			   h = _timingGain[3].a2 + a0/(1.+exp((eeV-_timingGain[3].x0)/
												_timingGain[3].dx));
			ampl[n] *= h;
		}

    } else { // No gain correction parameters found in CCF
		errstr<<"Gain correction parameters were not present for PN "
                "FF mode in the CCF. No correction will be applied "
			  <<Sas::warning("NoFFgainCorrectionFound");
    }

} // EpnEnergyCorrector::gainFfCorrect()

double EpnEnergyCorrector::reEmissionThreshold(double Eadu, int16 rawY) const
{
	double EkeV = Eadu * 1.e-3 * _adu_eV;

	return (_Acoef[0] + _Acoef[1] * EkeV) +
		   (_Bcoef[0] + _Bcoef[1] * EkeV) * (DBL(rawY)-_rawy0);
} // EpnEnergyCorrector::reEmissionThreshold()

std::vector<float> EpnEnergyCorrector::getRateDepGainCoeffs() const
{
 RETURN _rdc_a;
}

// Method: getRateDepPhaCoeffs - return the rate dependent PHS parameters
std::vector<double> EpnEnergyCorrector::getRateDepPhaCoeffs(double Ne, bool withGain) const
{
	vector<double> coeffs(9);

   // Return a single vector which contains the
   //     0-2  :  PHA constants
   //     3-5  :  Coefficients for this Ne and withGainTiming
   //     6-8  :  Errors on these coefficients

   // Timing and Burst modes are supported.
	unsigned short mode = calServer.state()->mode()->get();

   // Set the index - Timing=0, Burst=1
        int i=-1;
	if (mode == Xmm::FAST_TIMING) {
           i=0;
	} else if (mode == Xmm::FAST_BURST && _rdp_burst_support) {
           i=1;
        }

       // Interpolate the coefficients - depending on the withGain flag
        if (i>=0) {

	   if (withGain) {
	      coeffs[3] = (*_rdp_corry_1[i])(Ne);
  	      coeffs[4] = (*_rdp_corry_2[i])(Ne);
	      coeffs[5] = (*_rdp_corry_3[i])(Ne);
	      coeffs[6] = (*_rdp_ecorry_1[i])(Ne);
	      coeffs[7] = (*_rdp_ecorry_2[i])(Ne);
	      coeffs[8] = (*_rdp_ecorry_3[i])(Ne);
           } else {
              coeffs[3] = (*_rdp_corrn_1[i])(Ne);
	      coeffs[4] = (*_rdp_corrn_2[i])(Ne);
	      coeffs[5] = (*_rdp_corrn_3[i])(Ne);
	      coeffs[6] = (*_rdp_ecorrn_1[i])(Ne);
	      coeffs[7] = (*_rdp_ecorrn_2[i])(Ne);
	      coeffs[8] = (*_rdp_ecorrn_3[i])(Ne);
	   }

           coeffs[0] = _rdp_pha[i][0];
           coeffs[1] = _rdp_pha[i][1];
           coeffs[2] = _rdp_pha[i][2];
      
        } else {
	   errstr<<"Rate-dependent PHA correction is not defined for mode: "
                 <<mode<<" returning zeroes"
		 <<Sas::warning("UnsupportedMode");
   //
           for (unsigned i=0; i<coeffs.size(); ++i)
			coeffs[i]=0.0;

        }
	return coeffs;
} // EpnEnergyCorrector::getRateDepPhaCoeffs

void EpnEnergyCorrector::readRDPHAcoeffs(Table *tab)
{
    // Check for correct structure
        if (tab->hasColumn("NY_e")) {

          std::vector<double>     nye, corry_1, ecorry_1, // rate-dependent coefficients
                                          corry_2, ecorry_2, corry_3, ecorry_3,
                                          nne, corrn_1, ecorrn_1,
                                          corrn_2, ecorrn_2, corrn_3, ecorrn_3;

         // Read coefficient columns from the RDPHA table
          CalOalUtility::fillVector(tab, "NY_e", nye);
          CalOalUtility::fillVector(tab, "CORRY_1", corry_1);
          CalOalUtility::fillVector(tab, "ECORRY_1", ecorry_1);
          CalOalUtility::fillVector(tab, "CORRY_2", corry_2);
          CalOalUtility::fillVector(tab, "ECORRY_2", ecorry_2);
          CalOalUtility::fillVector(tab, "CORRY_3", corry_3);
          CalOalUtility::fillVector(tab, "ECORRY_3", ecorry_3);
          CalOalUtility::fillVector(tab, "NN_e", nne);
          CalOalUtility::fillVector(tab, "CORRN_1", corrn_1);
          CalOalUtility::fillVector(tab, "ECORRN_1", ecorrn_1);
          CalOalUtility::fillVector(tab, "CORRN_2", corrn_2);
          CalOalUtility::fillVector(tab, "ECORRN_2", ecorrn_2);
          CalOalUtility::fillVector(tab, "CORRN_3", corrn_3);
          CalOalUtility::fillVector(tab, "ECORRN_3", ecorrn_3);

         // Make interpolation curves
          _rdp_corry_1.push_back(new Slatec::LinearInterpolant<double>(nye, corry_1));
          _rdp_corry_2.push_back(new Slatec::LinearInterpolant<double>(nye, corry_2));
          _rdp_corry_3.push_back(new Slatec::LinearInterpolant<double>(nye, corry_3));
          _rdp_corrn_1.push_back(new Slatec::LinearInterpolant<double>(nne, corrn_1));
          _rdp_corrn_2.push_back(new Slatec::LinearInterpolant<double>(nne, corrn_2));
          _rdp_corrn_3.push_back(new Slatec::LinearInterpolant<double>(nne, corrn_3));
          _rdp_ecorry_1.push_back(new Slatec::LinearInterpolant<double>(nye, ecorry_1));
          _rdp_ecorry_2.push_back(new Slatec::LinearInterpolant<double>(nye, ecorry_2));
          _rdp_ecorry_3.push_back(new Slatec::LinearInterpolant<double>(nye, ecorry_3));
          _rdp_ecorrn_1.push_back(new Slatec::LinearInterpolant<double>(nne, ecorrn_1));
          _rdp_ecorrn_2.push_back(new Slatec::LinearInterpolant<double>(nne, ecorrn_2));
          _rdp_ecorrn_3.push_back(new Slatec::LinearInterpolant<double>(nne, ecorrn_3));

         // Read the PHA constants from table attributes
          std::vector<double> phas;
          phas.push_back(tab->attribute("PHA_1")->asReal());
          phas.push_back(tab->attribute("PHA_2")->asReal());
          phas.push_back(tab->attribute("PHA_3")->asReal());
          _rdp_pha.push_back(phas);

       } else {
          errstr<<"Column NY_e not found in block RDPHA_DERIV of CCF "
                <<"EPN_CTI_nnnn.CCF "
                <<"please update your CCF"<<Sas::error("UnexpectedCcfData");
       }

} //EpnEnergyCorrector::ReadRDPHAcoeffs
 
EpnEnergyCorrector::~EpnEnergyCorrector()
{
	//	free up memory
	for (unsigned n=0; n<_gainParams.size(); ++n) {
		delete _gainParams[n]; _gainParams[n] = 0;
	}
	_gainParams.resize(0);
	_gainParamsSuppl.resize(0);
	_gainParamsSet = false;
} // EpnEnergyCorrector::~EpnEnergyCorrector()


//
//	---------------------( RGS )--------------------------------------------
//
RgsEnergyCorrector::RgsEnergyCorrector(
			Ccd ccd,
			const CcfConstituentTable &aduConvTab,
			const Table *darkFrameTable,
			const Table *offsetOverrideTable,
			CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent),
	  EnergyCorrector(ccd, aduConvTab, parent)
{
	const Table *aduConv = aduConvTab.block();

	//
	//	read rejection/acceptance thresholds and charge pixel
	//	ratios
	//
	_qSplitR[0] = 1.f;	//	by definition
	_qSplitR[1] = aduConv->attribute("QSPLITR2")->asReal();
	_qSplitR[2] = aduConv->attribute("QSPLITR3")->asReal();
	_qSplitR[3] = aduConv->attribute("QSPLITR4")->asReal();

	ColumnData *cdRejThresh	= aduConv->column("REJTHRESH")->data(),
			   *cdAccThresh = aduConv->column("ACCTHRESH")->data();
	real32	*rejThresh = cdRejThresh->real32Data(),
			*accThresh = cdAccThresh->real32Data();

	unsigned n = 0, N = aduConv->rows();
	for (unsigned n=0; n<N; ++n) {
		if (n+1 == ccd) {
			unsigned nn = cdRejThresh->elements();
			for (unsigned j=0; j<nn; ++j) {
				unsigned p = cdRejThresh->position(n, j);
				//
				//	check for negative threshold values
				//
				if (rejThresh[p] < 0. || accThresh[p] < 0.)
					errstr<<"rejection/acceptance threshold for CCD #"
						  <<ccd<<"/node #"<<j
						  <<" is negative"
						  <<Sas::error("CcfDataIntegrityError");
				_rejectionThresholds.push_back(rejThresh[p]);
				_acceptanceThresholds.push_back(accThresh[p]);
			}
			break;
		}
	}
	if (n>=N)
		errstr<<"Did not find rejection/acceptance tresholds for RGS CCD #"
			  <<ccd<<" in `"<<aduConv->qualifiedName()<<'\''
			  <<Sas::error("CcfDataIntegrityError");

	//
	//	try to read RAND_MIN/RAND_MAX from attributes
	//
	_eRandMin = 0.;
	_eRandMax = 1.;	//	default values
	string rmin("RAND_MIN"),
		   rmax("RAND_MAX");
	if (aduConv->hasAttribute(rmin) &&
		aduConv->hasAttribute(rmax)) {
		_eRandMin = aduConv->attribute(rmin)->asReal();
		_eRandMax = aduConv->attribute(rmax)->asReal();
	} else
		if (aduConv->parent()->hasAttribute(rmin) &&
			aduConv->parent()->hasAttribute(rmax)) {
			_eRandMin = aduConv->parent()->attribute(rmin)->asReal();
			_eRandMax = aduConv->parent()->attribute(rmax)->asReal();
		} else
			errstr<<"Did not find "<<rmin<<'/'<<rmax<<" attributes in table "
				  <<aduConv->qualifiedName()<<"' or associated set - "
					"randomization will use default value range ["<<
					_eRandMin<<", "<<_eRandMax<<"]"
				  <<Sas::warning("AttributesNotFound");

	//
	//	read dark frame parameters
	//
	unsigned nNodes = xmmPayLoad.rgs1.nNodes();
	_a0.resize(nNodes);
	_a1.resize(nNodes);
	_a2.resize(nNodes);
	_a3.resize(nNodes);
	_b0.resize(nNodes);
	_b1.resize(nNodes);
	_b2.resize(nNodes);
	_b3.resize(nNodes);
	ColumnData *cda0 = darkFrameTable->column("A_0")->data(),
			   *cda1 = darkFrameTable->column("A_1")->data(),
			   *cda2 = darkFrameTable->column("A_2")->data(),
			   *cda3 = darkFrameTable->column("A_3")->data(),
			   *cdb0 = darkFrameTable->column("B_0")->data(),
			   *cdb1 = darkFrameTable->column("B_1")->data(),
			   *cdb2 = darkFrameTable->column("B_2")->data(),
			   *cdb3 = darkFrameTable->column("B_3")->data(),
			   *cdccd = darkFrameTable->column("CCD_ID")->data(),
			   *cdnode= darkFrameTable->column("NODE_ID")->data();
	real32	*a0 = cda0->real32Data(),
			*a1 = cda1->real32Data(),
			*a2 = cda2->real32Data(),
			*a3 = cda3->real32Data(),
			*b0 = cdb0->real32Data(),
			*b1 = cdb1->real32Data(),
			*b2 = cdb2->real32Data(),
			*b3 = cdb3->real32Data();
	int8	*ccdid = cdccd->int8Data(),
			*nodeid = cdnode->int8Data();

	n = 0;
	N = darkFrameTable->rows();
	bool found = false;
	for (; n<N; ++n)
		if (ccdid[n] == ccd) {
			found = true;
			unsigned node = nodeid[n];
			if (node < nNodes) {
				_a0[node] = a0[n]; _a1[node] = a1[n];
				_a2[node] = a2[n]; _a3[node] = a3[n];
				_b0[node] = b0[n]; _b1[node] = b1[n];
				_b2[node] = b2[n]; _b3[node] = b3[n];
			} else
				errstr<<"Illegal node ID ("<<node<<") seen in `"
					  <<darkFrameTable->qualifiedName()
					  <<Sas::error("CcfDataIntegrityError");
		}

	if (!found)
		errstr<<"Did not find dark frame parameters for RGS CCD #"
			  <<ccd<<" in `"<<darkFrameTable->qualifiedName()<<"'"
			  <<Sas::error("CcfDataIntegrityError");

	delete cda0;
	delete cda1;
	delete cda2;
	delete cda3;
	delete cdb0;
	delete cdb1;
	delete cdb2;
	delete cdb3;
	delete cdccd;
	delete cdnode;

	//
	//	optionally read diagnostic offset data from dynamic CCF constituent
	//	created by rgsoffsetcalc
	//
	_hasDiagnosticOffsetData = offsetOverrideTable;
	if (!_hasDiagnosticOffsetData) {
		flatOffsetCurves();	//	define flat offset curves

		return;
	}

	ColumnData *cdtime = offsetOverrideTable->column("TIME")->data(),
			   *cdmean= offsetOverrideTable->column("MEAN")->data();
    cdccd = offsetOverrideTable->column("CCD_ID")->data(),
	cdnode= offsetOverrideTable->column("NODE_ID")->data();

	real64	*time = cdtime->real64Data();
	ccdid = cdccd->int8Data(),
	nodeid = cdnode->int8Data();
	real32	*mean = cdmean->real32Data();

	vector<double> t[nNodes],
				   offset[nNodes];
	double prevTime[nNodes];
	for(n=0; n<nNodes; prevTime[n++] = -1.) ;
	n = 0;
	N = offsetOverrideTable->rows();
	for (; n<N; ++n)
		if (UNT(ccdid[n]) == ccd) {
			unsigned node = nodeid[n];
			if (node < nNodes) {
				if (time[n] != prevTime[node]) {
					prevTime[node] = time[n];
					t[node].push_back(time[n]);
					offset[node].push_back(mean[n]);
				} else
					errstr<<"Time in row "<<n<<" of `"
						  <<offsetOverrideTable->qualifiedName()
						  <<"' for CCD "<<ccd<<" node "<<node<<" identical "
							"to previous time - ignoring entry"
						  <<Sas::warning("WrongTime");
			} else
				errstr<<"Illegal node ID ("<<node<<") seen in `"
					  <<darkFrameTable->qualifiedName()
					  <<Sas::error("CcfDataIntegrityError");
		}

	delete cdtime;
	delete cdmean;

	if (!t[0].size() && !t[1].size()) {
		errstr<<"Did not find entry for CCD #"<<ccd<<" in `"
			  <<offsetOverrideTable->qualifiedName()<<"' - offset correction "
				"for this CCD will use CCF-based values"
			  <<Sas::warning("NoDynamicOffsetData");
		unsigned nNodes = xmmPayLoad.rgs1.nNodes();
		for (unsigned node=0; node<nNodes; flatOffsetCurve(node++)) ;

		return;
	}

	for (unsigned node=0; node<nNodes; ++node) {
		if (t[node].size()) {
			//
			//	insert duplicate first/last offset value for three
			//	early/late times times such that the offset curve has a
			//	plateau at the beginning/end
			//
			t[node].insert(t[node].begin(), 100.);
			t[node].insert(t[node].begin(), 10.);
			t[node].insert(t[node].begin(), 0.);
			for (n=0; n<3; ++n)
				offset[node].insert(offset[node].begin(), offset[node][0]);

			double missionDuration = Xmm::missionEndTime -
											Xmm::missionReferenceTime;
			t[node].push_back(missionDuration-1000.);
			t[node].push_back(missionDuration-100.);
			t[node].push_back(missionDuration);
			for (n=0; n<3; ++n)
				offset[node].push_back(offset[node][offset[node].size()-1]);
			_offsetCurves[node] =
				new Slatec::CubicHermiteInterpolant<double>(t[node],
															offset[node]);
		} else {
			//	no data for node <node>
			errstr<<"No dynamic offset data for node "<<node<<" have been "
					"found - using flat offset curve"
				  <<Sas::warning("NoDynamicOffsetData");
			flatOffsetCurve(node);
		}
	}
} // RgsEnergyCorrector::RgsEnergyCorrector()


void RgsEnergyCorrector::flatOffsetCurve(unsigned node)
{
	if (node >= xmmPayLoad.rgs1.nNodes())
		errstr<<"Illegal node ID "<<node<<Sas::error("IllegalNodeId");

	//
	//	define flat offset curve
	//
	vector<double>	timeVec(2),
					offsetVec(2);
	timeVec[0] = 0.;
	timeVec[1] = 1000.;
	offsetVec[0] = offsetVec[1] = offsetValue(Xmm::Node(node));
	_offsetCurves[node] =
			new Slatec::CubicHermiteInterpolant<double>(timeVec, offsetVec);
} // RgsEnergyCorrector::flatOffsetCurve()


void RgsEnergyCorrector::flatOffsetCurves()
{
	unsigned nNodes = xmmPayLoad.rgs1.nNodes();
	for (unsigned n=0; n<nNodes; flatOffsetCurve(n++)) ;
} // RgsEnergyCorrector::flatOffsetCurves()



CalReal32Vector &RgsEnergyCorrector::offsetCorrect(
												 const CalReal64Vector &time,
												 const CalInt16Vector &rawX,
												 const CalInt16Vector &rawY,
												 const CalInt8Vector &shape,
												 const CalInt16Vector &eIn,
												 CalReal32Vector &eOut) const
{
	unsigned N = eIn.size();
	CalInt8Vector nodes(0, N);
	int nodeIndex = parent()->node();
	for (unsigned n=0; n<N; nodes[n++] = nodeIndex) ;

	offsetCorrect(time, rawX, rawY, shape, nodes, eIn, eOut);

	return eOut;
} // RgsEnergyCorrector::offsetCorrect()


CalReal32Vector &RgsEnergyCorrector::offsetCorrect(
												 const CalReal64Vector &time,
												 const CalInt16Vector &rawX,
												 const CalInt16Vector &rawY,
												 const CalInt8Vector &shape,
												 const CalInt8Vector &nodes,
												 const CalInt16Vector &eIn,
												 CalReal32Vector &eOut) const
{
	randomize(eIn, eOut);
	float offset;
	char nNodes = _offsetConst.size();
	for (unsigned n=0; n<eIn.size(); ++n) {
		if (nodes[n] < nNodes)
			offset = (*_offsetCurves[nodes[n]])(time[n]);
		else {
			errstr<<"No offset value for node "<<INT(nodes[n])
				  <<" event #"<<n<<" - no correction applied"
				  <<Sas::warning("NoSuchValue");
			offset = 0.f;
		}

		eOut[n] -= shapeToGrade(shape[n]) * (offset +
					   darkCurrent(rawX[n], rawY[n], nodes[n]));
	}

	return eOut;
} // RgsEnergyCorrector::offsetCorrect()


//
//	HTR mode offset correction
//
CalReal32Vector &RgsEnergyCorrector::offsetCorrect(
												 const CalReal64Vector &time,
												 const CalInt16Vector &rawX,
												 const CalInt8Vector &nodes,
												 const CalInt16Vector &eIn,
												 CalReal32Vector &eOut) const
{
	CalState *state = calServer.state();

	//	verify we have HTR mode
	Mode mode = state->mode()->get();
	if (mode != Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD &&
		mode != Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD)
		errstr<<"Called RGS HTR-specific offset correction in mode `"
			  <<mode<<"'"<<Sas::error("ImproperMode");

	//	set OCB to 3 + construct dummy rawY/shape vectors
	//	set rawY to middle of readout window in OCB 3; shape is 0
	unsigned oldocb = state->ocb()->get();
	state->ocb()->set(3);

	CalInt16Vector rawY(0, rawX.size());
	int16 v = static_cast<int16>(calServer.instrument()->maxY()/3/2);
	rawY = v;
	CalInt8Vector shape(0, rawX.size());
	shape = 0;

	//	call spectroscopy offset correction
	CalReal32Vector &e =
					offsetCorrect(time, rawX, rawY, shape, nodes, eIn, eOut);
	state->ocb()->set(oldocb);

	return e;
} // RgsEnergyCorrector::offsetCorrect()


float RgsEnergyCorrector::darkCurrent(int16 rawX, int16 rawY, unsigned node)
const
{
	double x = DBL(rawX-1),
		   y = DBL(rawY-1),
		   dc = _a0[node]*exp(_a1[node]*x) +
				_a2[node]*x +
				.5*(_a3[node] + _b3[node]) +
				_b0[node]*exp(_b1[node]*y) +
				_b2[node]*y;	//	see [2], p. 47

	return dc;
} // RgsEnergyCorrector::darkCurrent()


double RgsEnergyCorrector::rejectionThreshold(unsigned grade) const
{
	return rejectionThreshold(grade, calServer.state()->node()->get());
} // RgsEnergyCorrector::rejectionThreshold()


double RgsEnergyCorrector::rejectionThreshold(unsigned grade,
											  Node node) const
{
	if (grade < 1 || grade > NGRADES)
		errstr<<"Passed grade value ("<<grade<<") invalid - must be in "
				"[1, "<<NGRADES<<"]"<<Sas::error("InvalidGradeValue");
	if (node < 0 || node > INT(xmmPayLoad.rgs1.nNodes()))
		errstr<<"Passed node value ("<<node<<") invalid - must be in "
				"[0, "<<xmmPayLoad.rgs1.nNodes()<<"]"
			  <<Sas::error("InvalidNode");

	return _rejectionThresholds[node] / _qSplitR[grade-1];
} // RgsEnergyCorrector::rejectionThreshold()


double RgsEnergyCorrector::acceptanceThreshold(unsigned grade) const
{
	return acceptanceThreshold(grade, calServer.state()->node()->get());
} // RgsEnergyCorrector::acceptanceThreshold()


double RgsEnergyCorrector::acceptanceThreshold(unsigned grade,
											   Node node) const
{
	if (grade < 1 || grade > NGRADES)
		errstr<<"Passed grade value ("<<grade<<") invalid - must be in "
				"[1, "<<NGRADES<<"]"<<Sas::error("InvalidGradeValue");
	if (node < 0 || node > INT(xmmPayLoad.rgs1.nNodes()))
		errstr<<"Passed node value ("<<node<<") invalid - must be in "
				"[0, "<<xmmPayLoad.rgs1.nNodes()<<"]"
			  <<Sas::error("InvalidNode");

	return _acceptanceThresholds[node] / _qSplitR[grade-1];
} // RgsEnergyCorrector::acceptanceThreshold()


vector<float> RgsEnergyCorrector::qSplitRatios() const
{
	vector<float> v;
	for (unsigned n=0; n<NGRADES; v.push_back(_qSplitR[n++])) ;
	return v;
} // RgsEnergyCorrector::qSplitRatios()



unsigned short RgsEnergyCorrector::shapeToGrade(unsigned short shape) const
{
	//
	//  SHAPE      GRADE
	//    0          1
	//    1          4
	//    2          2
	//    3          2
	//    4          3
	//    5          3
	//    6          3
	//    7          3
	//	 >7          2 + warning
	//
	static unsigned short trans[] = { 1, 4, 2, 2, 3, 3, 3, 3 };

	if (shape <NUM_OF(trans))
		return trans[shape];

	errstr<<"Shape code "<<shape<<" has no associated grade value - shape "
			"codes must be in interval (0, "<<NUM_OF(trans)-1
		  <<"] - returning with grade 2"
		  <<Sas::warning("IllegalNullShape");

	return 2;
} // RgsEnergyCorrector::shapeToGrade()


void RgsEnergyCorrector::randomize(const CalInt16Vector &eIn,
								   CalReal32Vector &eOut) const
{
	calServer.setRandom(_eRandMin, _eRandMax);
	eIn.toReal(eOut);
	calServer.setRandom();
} // RgsEnergyCorrector::randomize()


RgsEnergyCorrector::~RgsEnergyCorrector()
{
	unsigned nNodes = xmmPayLoad.rgs1.nNodes();
	for (unsigned n=0; n<nNodes; ++n)
		delete _offsetCurves[n];
} // RgsEnergyCorrector::~RgsEnergyCorrector()

