// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	HKvalRangesServer.cc --- implementation of HKvalRangesServer class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-10-19: 1.0.0: - first incarnation
//
//	AUTHOR:
//		U. Lammers, Tue Oct 19 12:12:00 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/HKvalRangesServer.cc,v 1.12 2009/07/06 14:25:19 aibarra Exp $
//

#include <algorithm>

#include "stringutil.h"
#include "CalOalUtility.h"
#include "HKvalRangesServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( HKvalRanges )-------------------------------------------------------
//
HKvalRanges::MapType::const_iterator HKvalRanges::has(const string &key) const
{
	return find(uppercase(key));
} // HKvalRanges::has()


string HKvalRanges::get(MapType::const_iterator where,
						const string &key) const
{
	if (where != end())
		return (*where).second;

	errstr<<"There is no validity range associated with parameter `"
		  <<key<<"'"
		  <<Sas::error("NoSuchData");

	return "";
} // HKvalRanges::get()


string HKvalRanges::get(const string &key) const
{
	return get(find(uppercase(key)), key);
} // HKvalRanges::get()


string HKvalRanges::getExpanded(MapType::const_iterator where,
								const string &key) const
{
	return expand(get(where, key), (*where).first);
} // HKvalRanges::getExpanded()


string HKvalRanges::getExpanded(const string &key) const
{
	return expand(get(key), key);
} // HKvalRanges::getExpanded()


string HKvalRanges::expand(const string &expr, const string &key) const
{
	string s(expr);
	string::size_type p;
	while ((p = s.find('@')) != string::npos)
		s.replace(p, 1, key);

	return s;
} // HKvalRanges::expand()


HKvalRanges HKvalRanges::reduceWithShortNames(const vector<string> &
																shortHKnames)
{
	HKvalRanges rlist;
	for (MIter i = begin(); i != end(); ++i) {
		string parName = (*i).first;
		MCIter j = _translationMap.find(parName);
		if (j != _translationMap.end())
			//	yes, a long name for which there is a short name replacement
			parName = (*j).second;

		//	see if the name occurs in the the list of HK column names
		//	if yes it has to be copied in the the new list
		vector<string>::const_iterator p = ::find(shortHKnames.begin(),
												  shortHKnames.end(),
												  parName);
		if (p != shortHKnames.end()) {
			//	yup, found one
			rlist[parName] = (*i).second;
			erase(i);
		}
	}

	return rlist;
} // HKvalRanges::reduceWithShortNames()


string HKvalRanges::selectionExpression(bool translateLongToShort) const
{
	string s;
	MapType::const_iterator i = begin();
	while (i != end()) {
		s += '(' + getExpanded(i++) + ')';
		if (i != end())
			s += "&&";
	}

	//	translate long names into short names
	if (translateLongToShort)
		for (MapType::const_iterator i = _translationMap.begin();
			 i != _translationMap.end(); ++i) {
			string::size_type p;
			while ((p = s.find((*i).first)) != string::npos)
				s.replace(p, (*i).first.length(), (*i).second);
		}

	return s;
} // HKvalRanges::selectionExpression()


//
//	---( HKvalRangesServer )---------------------------------------------------
//
/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<HKvalRangesServer>::CalAtomCollection(
					const XmmInstrument *instrument,
					const Ccf *ccf,
					Xmm::Filter filter,
					AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_collection	= new vector<HKvalRangesServer *>;
	_collection->push_back(new HKvalRangesServer);
	HKvalRangesServer &hkval = *(*_collection)[0];

	CifEntry *e = ccf->entry(instrument, Ccf::HKParmInt);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet,
							e->queryBlockName(Ccf::toString(Ccf::HKParmInt)));
	const Table *table = ccftab.block();

	errstr<<"Reading HK validity ranges from file "<<e->fileName()
		  <<Sas::message(Msg::Lib, Msg::Verbose);


	//	read all parameters corresponding to the current instrument, the
	//	relevant telescope and the satellite
	ColumnData *parmName = table->column("PARM_NAME")->data(),
			   *parmRange = table->column("PARM_RANGE")->data();
	for (unsigned n=0; n<table->rows(); ++n) {
		string name = uppercase(stringColumnRow(parmName, n)),
			   range = uppercase(stringColumnRow(parmRange, n));
		hkval[name] = range;

		if (errHandler.verbosity(Msg::Lib) >= Msg::Verbose)
			errstr<<name<<" : "<<range<<endl
				  <<Sas::message(Msg::Lib, Msg::Verbose);
	}

	delete parmName;
	delete parmRange;

	//
	//	read name translation table if present
	//
	if (dataSet->hasBlock("TRANSLATION")) {
		CcfConstituentTable ccftab(dataSet, "TRANSLATION");
		const Table *table = ccftab.block();

		ColumnData *shortName = table->column("SHORT_NAME")->data(),
				   *longName = table->column("LONG_NAME")->data();

		HKvalRanges::MapType map;
		for (unsigned n=0; n<table->rows(); ++n) {
			string shortHK = uppercase(stringColumnRow(shortName, n)),
				   longHK = uppercase(stringColumnRow(longName, n));

			map[longHK] = shortHK;

			if (errHandler.verbosity(Msg::Lib) >= Msg::Verbose)
				errstr<<shortHK<<" => "<<longHK<<endl
					  <<Sas::message(Msg::Lib, Msg::Verbose);
		}
		hkval.translationMap(map);
	}

	dataSetServer->close(dataSet);
} // CalAtomCollection<HKvalRangesServer>::CalAtomCollection()


void HKvalRangesServer::dump(ostream &os)
{
	os<<"HKvalRangesServer{\n";
	for (const_iterator i = begin(); i != end(); ++i)
		os<<(*i).first<<"="<<(*i).second<<endl;
	os<<"}";
} // HKvalRangesServer::dump()

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
HKvalRangesServer *CalAtomCollection<HKvalRangesServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // HKvalRangesCollection<HKvalRangesServer>::select()


HKvalRanges &HKvalRangesServer::valRanges()
{
	return *this;
} // HKvalRangesCollection::valRanges;

