// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OmFluxConverter --- class which converts raw count rates
//
//	DESCRIPTION:
//		Instances of class OmFluxConverter can
//			a) correct raw detector count rate for instrumental effects
//			b) provide a distinct set of parameters for the frame time
//			   calculation
//		For details see [2]
//
//	CHANGE HISTORY:
//		UL, 2000-01-27: 1.0.0: - initial version
//
//	AUTHOR:
//		U. Lammers (UL), Thu Jan 27 11:23:42 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, XMM instrument team
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/OmFluxConverter.h,v 1.9 2007/03/02 14:54:29 rsaxton Exp $
//

#ifndef OM_FLUX_CONVERTER_H
#define OM_FLUX_CONVERTER_H

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"

//
//	---( OmFluxConverter )-------------------------------------------------
//
class OmFluxConverter : public CalAtom,
	  public virtual CalChild<CalAtomCollection<OmFluxConverter> > {
public:
	struct FrameParameters {
		ValueError	nHorizontalPixels,
					nVerticalPixels,
					verticalTransferTime,
					horizontalTransferTime,
					nVerticalPixelsImage,
					verticalHorizontalSwitchTime,
					horizontalVerticalSwitchTime,
					redundantCcdOffset;
	};
	OmFluxConverter(const CcfConstituentTable &photToNatTab,
					CalAtomCollection<OmFluxConverter> *parent = 0);
	OmFluxConverter(const CcfConstituentTable &photToNatTab,
					const CcfConstituentTable &coincidenceTab,
					CalAtomCollection<OmFluxConverter> *parent = 0);
	OmFluxConverter(const CcfConstituentTable &photToNatTab,
					const CcfConstituentTable &coincidenceTab,
					const std::vector<double> &degradCoeffs,
					//const CcfConstituentTable &degradationTab,
					CalAtomCollection<OmFluxConverter> *parent = 0);
		//
		//	PURPOSE:
		//		Construct OmFluxConverter object.
		//
		//	ARGUMENTS:
		//		frameParameters	: CCD parameters associated with frame time
		//						  calculation
		//		phdFrac			: fraction of detected events as measured at
		//						  the time of most recent photometric
		//						  calibration
		//		thresh			: threshold value
		//


	static unsigned id() { return Ccf::PhotToNat; }
		//
		//	PURPOSE:
		//		Return a uniue ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	virtual void dump(std::ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	virtual FrameParameters frameParameters() const;
		//
		//	PURPOSE:
		//		Return frame parameters associated with frame time
		//		calculation
		//


	virtual double phdFrac() const;
	virtual double thresh() const;
		//
		//	PURPOSE:
		//		Return phdFrac/thresh parameters.
		//

	virtual vector<double> degradationCoeffs() const;
        //
        //  PURPOSE:
        //      Return degradation coefficients.
        //

	virtual ValueAsymmetricError fluxCorrect(
							ValueAsymmetricError rawRate,
							double frameTime,
							double deadFrac,
							double eventLoss = 0,
							bool extendedSource = false,
							bool empiricalLinearityCorrection = false) const;
		//
		//	PURPOSE:
		//		Correct a raw count rate for effects of CCD deadtime and
		//		detector non-linearity; please note: for non-extended
		//		sources the linearity correction will fail if <rawRate>
		//		is greater than or equal to the frameRate (=1/frameTime);
		//		this condition is detected, a warning will be issued and
		//		the largest positive real number returned to the caller.
		//
		//	ARGUMENTS:
		//		rawRate		: raw input rate [count/PSF-width/s] + associated
		//					  asymmetric uncertainties
		//		frameTime	: CCD frame time [s]
		//		deadFrac	: deadtime caused by frame transfer expressed
		//					  as fraction of CCD frame time
		//		eventLoss	: fraction of events below detection threshold in
		//					  the most recent PHD
		//		extendedSource:
		//					  boolean switch determining whether the
		//					  celestial source is extended or not; in the
		//					  former case the coinidence loss correction is
		//					  NOT applied (see [2] for more details)
		//		empiricalLinearityCorrection:
		//					  if set to true the count rate is, in addition
		//					  to the theoretical linearity correction,
		//					  corrected by an empirical factor (see [2]
		//					  for details)
		//
		//	RETURNS:
		//		Count rate corrected for instrumental effects [cts/s] +
		//		associated asymmetric uncertainties
		//


	virtual ~OmFluxConverter();
		//
		//	PURPOSE:
		//		Destruct OmFluxConverter object
		//

private:
	//	data members
	FrameParameters	_frameParameters;
	double			_phdFrac,
					_thresh;

	std::vector<double>	_linearityCoeff;
	std::vector<double>	_degradCoeffs;

	//	methods
    void init(const CcfConstituentTable &, const CcfConstituentTable *tab = 0);
	//void init(const CcfConstituentTable &, const CcfConstituentTable *tab = 0,
                                           //const std::vector<double> *dvec = 0);
};

#endif
